"""
Determine what compiler vendor was used to compile a given binary, by checking the .comments ELF section
"""
from typing import Iterable
from enum import IntEnum
from pathlib import Path
from elftools.elf.elffile import ELFFile
from elftools.common.exceptions import ELFError
from e4s_cl.logger import get_logger
from e4s_cl.util import which

LOGGER = get_logger(__name__)


def _gnu_check(string: str) -> bool:
    return 'GCC' in string


def _llvm_check(string: str) -> bool:
    return 'clang' in string


def _intel_check(_: str) -> bool:
    return False


def _amd_check(string: str) -> bool:
    return 'AMD' in string


def _pgi_check(_: str) -> bool:
    return False


def _armclang_check(_: str) -> bool:
    return False


def _fujitsu_check(_: str) -> bool:
    return False


class CompilerVendor(IntEnum):
    """Enum with values describing compiler vendors"""
    GNU = 0
    LLVM = 1
    INTEL = 2
    AMD = 3
    PGI = 4
    ARMCLANG = 5
    FUJITSU = 6


CHECKS = {
    CompilerVendor.GNU: _gnu_check,
    CompilerVendor.LLVM: _llvm_check,
    CompilerVendor.INTEL: _intel_check,
    CompilerVendor.AMD: _amd_check,
    CompilerVendor.PGI: _pgi_check,
    CompilerVendor.ARMCLANG: _armclang_check,
    CompilerVendor.FUJITSU: _fujitsu_check
}

VENDOR_BINARIES = {
    CompilerVendor.GNU: ('gcc', 'g++', 'gfortran'),
    CompilerVendor.INTEL: ('icc', 'icpc', 'ifort'),
    CompilerVendor.PGI: ('pgcc', 'pgc++', 'pgfortran'),
    # We need AMD here
    CompilerVendor.LLVM: ('clang', 'clang++', 'flang'),
    CompilerVendor.ARMCLANG: ('armclang', 'armclang++', 'armflang'),
    CompilerVendor.FUJITSU: ('fcc', 'FCC', 'frt'),
}

# ROCm-compiled binaries contained 'AMD', 'clang' and 'GCC'
# Establishing an order for the checks is a simple way
# of ensuring the right value is returned
PRECENDENCE = [CompilerVendor.AMD, CompilerVendor.LLVM, CompilerVendor.GNU]


def _get_comment(elf_file: Path) -> str:
    """
    Returns the contents of the .comment sections of the ELF file passed as an argument
    """
    try:
        with open(elf_file, 'rb') as data:
            elf = ELFFile(data)
            comment_sections = filter(lambda x: x.name == '.comment',
                                      elf.iter_sections())
            return ' - '.join(
                map(lambda x: x.data().decode(), comment_sections))
    except (PermissionError, FileNotFoundError, IsADirectoryError,
            ELFError) as err:
        LOGGER.debug("Error reading comments of file %s: %s", str(elf_file),
                     str(err))
        return ''


def compiler_vendor(elf_file: Path) -> int:
    """
    Returns a value from CompilerVendor according to the contents of the .comment section of a binary
    """
    comment = _get_comment(elf_file)

    for vendor in PRECENDENCE:
        check = CHECKS.get(vendor)
        if check and check(comment):
            return vendor

    # By default, return GNU
    return CompilerVendor.GNU


def available_compilers() -> Iterable[int]:
    """Return a list of compiler identifiers for compilers found on the system"""

    available = set()
    for vendor, requirements in VENDOR_BINARIES.items():
        # Check which found every requirement for the compiler
        if len(set(filter(None, map(which,
                                    requirements)))) == len(requirements):
            available.add(vendor)

    return available



import os
from pathlib import Path
from tempfile import NamedTemporaryFile
from shlex import split
from sotools.linker import resolve
import tests
from e4s_cl.util import which
from e4s_cl.variables import ParentStatus
from e4s_cl.model.profile import Profile
from e4s_cl.cli.commands.profile.detect import (filter_files, COMMAND)
from e4s_cl.model.profile import Profile

__TEST_LIBRARY_SONAME__ = "libmpi.so"
__TEST_LIBRARY_NON_STANDARD__ = tests.ASSETS / "libgver.so.0"
__BLACKLISTED_FILES__ = {
    "/dev/null", "/proc/self/fd/0", "/root", "/root/file", "/etc/ld.so.cache"
}
__TEST_BINARY__ = f"file"
__TEST_ARGS__ = f"{resolve(__TEST_LIBRARY_SONAME__)}"


class ProfileDetectTest(tests.TestCase):

    def tearDown(self):
        Profile.controller().unselect()
        self.resetStorage()

    @tests.skipIf(not resolve(__TEST_LIBRARY_SONAME__),
                  f"{__TEST_LIBRARY_SONAME__} not found on this system")
    def test_filter_files_lib(self):
        library = str(resolve(__TEST_LIBRARY_SONAME__))

        libraries, files = filter_files([Path(library)])

        self.assertIn(library, libraries)
        self.assertFalse(files)

    def test_filter_files_non_std_lib(self):
        libraries, files = filter_files([__TEST_LIBRARY_NON_STANDARD__])

        self.assertFalse(libraries)
        self.assertIn(__TEST_LIBRARY_NON_STANDARD__.as_posix(), files)

    def test_filter_files_existing_file(self):
        with NamedTemporaryFile(dir=os.getcwd()) as datafile:
            libraries, files = filter_files([Path(datafile.name)])

            self.assertFalse(libraries)
            self.assertIn(datafile.name, files)

    def test_filter_files_blacklisted(self):
        libraries, files = filter_files(map(Path, *__BLACKLISTED_FILES__))

        self.assertFalse(libraries)
        self.assertFalse(files)

    @tests.skipIf(not which(__TEST_BINARY__),
                  f"{__TEST_BINARY__} not found on this system")
    def test_profile_detect_new_profile(self):
        self.assertCommandReturnValue(
            0, COMMAND,
            split(
                f"-p __test_profile_detect {__TEST_BINARY__} {__TEST_ARGS__}"))

    @tests.skipIf(not which(__TEST_BINARY__),
                  f"{__TEST_BINARY__} not found on this system")
    def test_profile_detect_existing_profile(self):
        profile = Profile.controller().create(
            {'name': '__test_profile_detect'})

        self.assertCommandReturnValue(
            0, COMMAND,
            split(
                f"-p __test_profile_detect {__TEST_BINARY__} {__TEST_ARGS__}"))

    @tests.skipIf(not which(__TEST_BINARY__),
                  f"{__TEST_BINARY__} not found on this system")
    def test_profile_detect_selected_profile(self):
        profile = Profile.controller().create(
            {'name': '__test_profile_detect'})
        Profile.controller().select(profile)

        self.assertCommandReturnValue(
            0, COMMAND, split(f"{__TEST_BINARY__} {__TEST_ARGS__}"))

    @tests.skipIf(not which(__TEST_BINARY__),
                  f"{__TEST_BINARY__} not found on this system")
    def test_profile_detect_no_profile(self):
        self.assertNotCommandReturnValue(
            0, COMMAND, split(f"{__TEST_BINARY__} {__TEST_ARGS__}"))

    @tests.skipIf(not which(__TEST_BINARY__),
                  f"{__TEST_BINARY__} not found on this system")
    def test_profile_detect_child(self):
        with ParentStatus():
            self.assertCommandReturnValue(
                0, COMMAND, split(f"{__TEST_BINARY__} {__TEST_ARGS__}"))

    def test_profile_detect_error(self):
        self.assertNotCommandReturnValue(
            0,
            COMMAND,
            split("python -c 'import sys; sys.exit(1)'"),
        )

    def test_profile_detect_no_command(self):
        self.assertNotCommandReturnValue(0, COMMAND, [])



"""Tests for e4s_cl.scripts.completion — static bash completion generator.

Coverage:
  CompletionGeneratorTest   tests generate_bash() and generate_completion_tree()
  CompletionInstallTest     tests regenerate_bash(), main(), _load_profile_names()
  CompletionFreshnessTest   tests that committed completions/e4s-cl.bash is current
"""

import io
import json
import os
import subprocess
import sys
import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch

import tests
import e4s_cl.scripts.completion as comp


# ── Tiny synthetic CLI tree (no e4s_cl import needed for logic tests) ─────────
_SIMPLE_TREE = {
    'name': 'root',
    'subcommands': [
        {
            'name': 'launch',
            'options': [
                {'names': ['--profile', '-p'], 'arguments': 1,
                 'expected_type': 'defined_profile'},
                {'names': ['--image'], 'arguments': 1},
                {'names': ['--verbose'], 'arguments': 0},
            ],
        },
        {
            'name': 'profile',
            'subcommands': [
                {
                    'name': 'select',
                    'positionals': [
                        {'arguments': 1, 'expected_type': 'defined_profile'}
                    ],
                    'options': [{'names': ['--help', '-h'], 'arguments': 0}],
                },
                {
                    'name': 'list',
                    'options': [{'names': ['--help', '-h'], 'arguments': 0}],
                },
            ],
        },
        {'name': '__hidden', 'options': []},
    ],
    'options': [
        {'names': ['--debug'], 'arguments': 0},
        {'names': ['--from'], 'arguments': 1,
         'values': ['openmpi', 'mpich'], 'expected_type': ''},
    ],
}


class CompletionGeneratorTest(tests.TestCase):
    """Tests for the bash-script generator (no filesystem or e4s_cl import)."""

    def _bash(self, tree=None):
        return comp.generate_bash(tree or _SIMPLE_TREE)

    def test_generate_bash_returns_string(self):
        out = self._bash()
        self.assertIsInstance(out, str)
        self.assertGreater(len(out), 100)

    def test_generated_bash_syntax_valid(self):
        result = subprocess.run(
            ['bash', '-n'],
            input=self._bash().encode(),
            capture_output=True,
        )
        self.assertEqual(result.returncode, 0,
                         f'bash -n failed:\n{result.stderr.decode()}')

    def test_generated_bash_has_complete_line(self):
        out = self._bash()
        self.assertIn('complete -F _e4s_cl_comp', out)
        self.assertIn('e4s-cl', out)

    def test_generated_bash_has_main_function(self):
        self.assertIn('_e4s_cl_comp() {', self._bash())

    def test_generated_bash_has_profile_helper(self):
        self.assertIn(comp._PROFILE_FN + '()', self._bash())

    def test_generated_bash_has_subcommands(self):
        out = self._bash()
        self.assertIn('launch', out)
        self.assertIn('profile', out)

    def test_generated_bash_hides_dunder_commands(self):
        out = self._bash()
        # __hidden must not appear in any compgen -W word-list
        for line in out.splitlines():
            if 'compgen' in line:
                self.assertNotIn('__hidden', line)

    def test_generated_bash_has_options(self):
        out = self._bash()
        self.assertIn('--debug', out)
        self.assertIn('--image', out)
        self.assertIn('--profile', out)

    def test_generated_bash_choices_inlined(self):
        # --from has choices ['openmpi', 'mpich'] — they must appear in the
        # generated case block for --from
        out = self._bash()
        self.assertIn('openmpi', out)
        self.assertIn('mpich', out)

    def test_generated_bash_profile_option_calls_helper(self):
        out = self._bash()
        # --profile and -p should trigger the profile-completion helper
        self.assertIn(comp._PROFILE_FN, out)

    def test_generate_completion_tree_returns_dict(self):
        tree = comp._generate_completion_tree()
        self.assertIsInstance(tree, dict)
        self.assertIn('subcommands', tree)

    def test_generate_completion_tree_has_launch(self):
        tree = comp._generate_completion_tree()
        names = [s['name'] for s in tree.get('subcommands', [])]
        self.assertIn('launch', names)

    def test_generate_completion_tree_has_profile_subcommands(self):
        tree = comp._generate_completion_tree()
        subs = {s['name']: s for s in tree.get('subcommands', [])}
        profile_subs = [s['name'] for s in subs.get('profile', {}).get('subcommands', [])]
        self.assertIn('list', profile_subs)
        self.assertIn('select', profile_subs)
        self.assertIn('create', profile_subs)

    def test_generate_completion_tree_idempotent(self):
        """Calling twice must not raise (tests shallow-copy fix for _COMMANDS)."""
        tree1 = comp._generate_completion_tree()
        tree2 = comp._generate_completion_tree()
        self.assertEqual(tree1, tree2)


class CompletionInstallTest(tests.TestCase):
    """Tests for regenerate_bash(), main(), and _load_profile_names()."""

    def setUp(self):
        self._tmpdir = tempfile.mkdtemp()
        self._orig_bash_file = comp._BASH_FILE
        self._orig_db_path = comp._DB_PATH
        comp._BASH_FILE = Path(self._tmpdir) / 'completions' / 'e4s-cl.bash'
        comp._DB_PATH = Path(self._tmpdir) / 'user.json'

    def tearDown(self):
        comp._BASH_FILE = self._orig_bash_file
        comp._DB_PATH = self._orig_db_path
        import shutil
        shutil.rmtree(self._tmpdir, ignore_errors=True)

    # ── regenerate_bash ────────────────────────────────────────────────────────

    def test_regenerate_creates_bash_file(self):
        comp.regenerate_bash()
        self.assertTrue(comp._BASH_FILE.exists())

    def test_regenerate_writes_valid_bash(self):
        comp.regenerate_bash()
        result = subprocess.run(
            ['bash', '-n', str(comp._BASH_FILE)],
            capture_output=True,
        )
        self.assertEqual(result.returncode, 0)

    def test_regenerate_idempotent(self):
        comp.regenerate_bash()
        content1 = comp._BASH_FILE.read_text()
        comp.regenerate_bash()
        content2 = comp._BASH_FILE.read_text()
        self.assertEqual(content1, content2)

    # ── main() ────────────────────────────────────────────────────────────────

    def test_main_regenerate_flag_creates_file(self):
        comp.main(['--regenerate'])
        self.assertTrue(comp._BASH_FILE.exists())

    def test_main_no_args_prints_source_when_file_exists(self):
        comp.regenerate_bash()
        out = io.StringIO()
        with patch('sys.stdout', out):
            comp.main([])
        self.assertIn('source', out.getvalue())
        self.assertIn('e4s-cl.bash', out.getvalue())

    def test_main_no_args_stderr_when_file_missing(self):
        err = io.StringIO()
        with patch('sys.stderr', err):
            comp.main([])
        self.assertIn('not found', err.getvalue().lower())

    # ── _load_profile_names ────────────────────────────────────────────────────

    def test_load_profile_names_empty_when_db_missing(self):
        names = comp._load_profile_names()
        self.assertEqual(names, [])

    def test_load_profile_names_reads_profiles(self):
        db_data = {'Profile': {'1': {'name': 'myprofile'}, '2': {'name': 'other'}}}
        comp._DB_PATH.write_text(json.dumps(db_data))
        names = comp._load_profile_names()
        self.assertIn('myprofile', names)
        self.assertIn('other', names)


class CompletionFreshnessTest(tests.TestCase):
    """Ensure the committed completions/e4s-cl.bash matches what the generator
    would produce today.  Fails loudly if a developer changes the CLI without
    regenerating the completion script.

    This test mirrors the CI workflow in .github/workflows/completion.yml.
    """

    def test_committed_bash_is_up_to_date(self):
        repo_file = Path(__file__).parent.parent / 'completions' / 'e4s-cl.bash'
        if not repo_file.exists():
            self.skipTest('completions/e4s-cl.bash not present in repo')
        tree = comp._generate_completion_tree()
        current = comp.generate_bash(tree)
        committed = repo_file.read_text()
        self.assertEqual(
            current, committed,
            'completions/e4s-cl.bash is out of date.\n'
            'Regenerate it with:\n'
            '   python scripts/generate_completion_bash.py > completions/e4s-cl.bash\n'
            'and commit the result.',
        )



"""
Modify the profile associated to the name passed as an argument.

Passing a value to the options **--name**, **--backend**, \
**--image**, **--source** will overwrite the profile's corresponding field.

Passing a value to **--add-files**, **--remove-files**, **--add-libraries**, \
**--remove-libraries** will add or remove elements from the list \
of files or libraries, accordingly.

The name argument can be omitted, in which case the selected profile is modified.
"""

from pathlib import Path
from e4s_cl import EXIT_SUCCESS, EXIT_FAILURE
from e4s_cl.error import UniqueAttributeError
from e4s_cl.cli import arguments
from e4s_cl.logger import get_logger
from e4s_cl.cli.cli_view import EditCommand
from e4s_cl.model.profile import Profile
from e4s_cl.cf.containers import EXPOSED_BACKENDS

LOGGER = get_logger(__name__)


class ProfileEditCommand(EditCommand):
    """``profile edit`` subcommand."""

    def _construct_parser(self):
        usage = f"{self.command} <profile_name> [arguments]"
        parser = arguments.get_model_identifier(self.model,
                                                prog=self.command,
                                                usage=usage,
                                                description=self.summary)

        parser.add_argument('--name',
                            help="change the profile's name",
                            metavar='<name>',
                            dest='name',
                            default=arguments.SUPPRESS)

        parser.add_argument(
            '--backend',
            help="change the profile's container technology." +
            f" Available backends are: {', '.join(EXPOSED_BACKENDS)}",
            metavar='<backend>',
            dest='backend',
            default=arguments.SUPPRESS)

        parser.add_argument('--image',
                            help="change the profile's image",
                            metavar='<path/to/image>',
                            dest='image',
                            type=str,
                            default=arguments.SUPPRESS)

        parser.add_argument('--source',
                            help="change the profile's setup script",
                            metavar='<path/to/script>',
                            dest='source',
                            type=arguments.posix_path,
                            default=arguments.SUPPRESS)

        parser.add_argument('--add-files',
                            help="Add files to the profile. Supports host:container syntax.",
                            metavar='<file>',
                            nargs='+',
                            type=arguments.posix_path,
                            default=arguments.SUPPRESS)

        parser.add_argument('--remove-files',
                            help="Remove files from the profile",
                            metavar='<file>',
                            nargs='+',
                            type=arguments.posix_path,
                            default=arguments.SUPPRESS)

        parser.add_argument('--add-libraries',
                            help="Add libraries to the profile",
                            metavar='<library>',
                            nargs='+',
                            type=arguments.posix_path,
                            default=arguments.SUPPRESS)

        parser.add_argument('--remove-libraries',
                            help="Remove libraries from the profile",
                            metavar='<library>',
                            nargs='+',
                            type=arguments.posix_path,
                            default=arguments.SUPPRESS)

        parser.add_argument('--wi4mpi',
                            help="Root of the WI4MPI installation to use",
                            metavar='<path>',
                            type=arguments.posix_path,
                            default=arguments.SUPPRESS)
        return parser

    def _parse_add_args(self, args, prof):
        added = set()
        for arg, attr in [('add_files', 'files'),
                          ('add_libraries', 'libraries')]:
            names = getattr(args, arg, [])
            for file_name in [Path(n).as_posix() for n in names]:
                if file_name and file_name not in prof.get(attr, []):
                    added.add(file_name)
                    prof[attr] = prof.get(attr, []) + [file_name]
                else:
                    LOGGER.error("File %s already in profile's %s", file_name,
                                 attr)

        return added

    def _parse_remove_args(self, args, prof):
        removed = set()
        for arg, attr in [('remove_files', 'files'),
                          ('remove_libraries', 'libraries')]:
            names = getattr(args, arg, [])
            for file_name in [Path(n).as_posix() for n in names]:
                if file_name and file_name in prof.get(attr, []):
                    removed.add(file_name)
                    prof[attr].remove(file_name)
                else:
                    LOGGER.error("File %s not in profile's %s", file_name,
                                 attr)

        return removed

    def main(self, argv):
        args = self._parse_args(argv)
        controller = Profile.controller()

        profile = args.profile
        profile_name = profile.get('name')

        updates = dict(profile)

        fields = {'name', 'backend', 'image', 'source', 'wi4mpi'}

        for field in fields:
            updates[field] = getattr(args, field, profile.get(field))

        for data in self._parse_add_args(args, updates):
            self.logger.info("Added %s to profile configuration '%s'.", data,
                             profile_name)

        for data in self._parse_remove_args(args, updates):
            self.logger.info("Removed %s from profile configuration '%s'.",
                             data, profile_name)

        try:
            controller.update(updates, {'name': profile_name})
        except UniqueAttributeError:
            LOGGER.error("Invalid parameters for edition: %s=%s",
                         Profile.key_attribute, updates[Profile.key_attribute])
            return EXIT_FAILURE

        return EXIT_SUCCESS


COMMAND = ProfileEditCommand(Profile, __name__)



"""
This command is intended to be run once for any given MPI library, and will \
create a :ref:`profile<profile>` to substitute that library in a container.

This is done by tracing the execution of a program using a given MPI library. \
The opened files and libraries will be detected, filtered, and stored in a profile.

It is highly encouraged to load the MPI library beforehand using the module \
system available (:code:`spack`/:code:`modules`/:code:`lmod`) to ensure the \
paths and dependencies are valid and loaded.

A :ref:`profile<profile>` name will be generated from the version of the found MPI \
library. Make sure it corresponds to the library you want to use, or continue to the \
below section.

Changing launcher and libraries
*********************************

**e4s-cl** can load an MPI library and run it without any other information. This is \
however a very generic operation that may fail on your system. The following \
options can be used to tune this process:

--mpi
    MPI installation to target. **e4s-cl** will search for a launcher and \
    libraries in this folder. If not supplied, the environment is used.

--launcher
    MPI launcher to use. Defaults to :code:`mpirun`.

--launcher_args
    Options to pass to the MPI launcher. Defaults to the empty string.

Alternatively, you can override the above options by providing a full command to \
run. This will require to compile an executable beforehand.

.. admonition:: The importance of inter-process communication

    This process relies on the execution of a sample MPI program to discover \
    its dependencies. In rare cases, a library will lazy-load network \
    libraries, preventing them from being detected with a simple example. \
    A message will appear in case some limitations were detected.

Examples
--------

Initializing using MPI resources from the environment:

.. code::

    module load mpich
    e4s-cl init  # Will use the launcher and library from the mpich module

Using :code:`srun` with special arguments:

.. code::

    module load cray-mpich-abi
    e4s-cl init --launcher srun
            --launcher_args "-A account-name -n 2 -N 2 -t 00:00:30"

Using a library installed on the system in :code:`/packages`:

.. code::

    e4s-cl init --mpi /packages/mpich --profile mpich

Using a fully formed command and an existing executable:

.. code-block:: bash

    $ e4s-cl init srun -n X -N Y ./executable

"""

import os
import json
import subprocess
import shlex
import argparse
from pathlib import Path
from typing import (
    Iterable,
    List,
    Optional,
    Union,
)
from e4s_cl import (
    E4S_CL_MPI_TESTER_SCRIPT_NAME,
    E4S_CL_SCRIPT,
    EXIT_FAILURE,
    EXIT_SUCCESS,
    INIT_TEMP_PROFILE_NAME,
)
from e4s_cl import logger, util
from e4s_cl.cf.detect_mpi import (
    filter_mpi_libs,
    library_install_dir,
    profile_mpi_name,
)
from e4s_cl.cf.trace import opened_files
from e4s_cl.cf.containers import guess_backend, EXPOSED_BACKENDS
from e4s_cl.cli.arguments import (
    REMAINDER,
    SUPPRESS,
    binary_in_path,
    get_parser,
    posix_path,
)
from e4s_cl.cli.command import AbstractCommand
from e4s_cl.cli.commands.profile.detect import COMMAND as detect_command
from e4s_cl.error import (UniqueAttributeError, ModelError)
from e4s_cl.model.profile import Profile

LOGGER = logger.get_logger(__name__)
_SCRIPT_CMD = os.path.basename(E4S_CL_SCRIPT)


def _check_mpirun(executable: Union[str, Path]):
    """
    Run hostname with the launcher and list the affected nodes
    """
    hostname_bin = util.which('hostname')
    if not hostname_bin:
        return

    with subprocess.Popen([executable, hostname_bin],
                          stdout=subprocess.PIPE) as proc:
        proc.wait()

        hostnames = {hostname.strip() for hostname in proc.stdout.readlines()}

    if len(hostnames) == 1:
        LOGGER.warning(
            "The target launcher %s uses a single host by default, "
            "which may tamper with the library discovery. Consider "
            "running `%s` using mpirun specifying multiple hosts.", executable,
            str(detect_command))


def _profile_from_args(args: argparse.Namespace) -> dict:
    """
    Create a dictionnary with all the profile related information passed as arguments
    """
    data = {}

    for attr in ['image', 'backend', 'source', 'wi4mpi']:
        value = getattr(args, attr, None)
        if value:
            data[attr] = value

    # Determine the backend if possible
    if data.get('image') and not data.get('backend'):
        data['backend'] = guess_backend(args.image)

    return data


def _analyze_binary(args: argparse.Namespace) -> int:
    """
    Trace a command to determine the dynamic dependencies of the MPI library
    """

    command = getattr(args, 'cmd', None)
    if not command:
        command = _generate_command(args)

    if not command:
        return EXIT_FAILURE

    # Run the program using the detect command and get a file list
    returncode = detect_command.main(command)

    if returncode != EXIT_SUCCESS:
        LOGGER.error("Tracing of MPI execution failed")
        return EXIT_FAILURE

    return EXIT_SUCCESS


def _find_tester() -> Optional[Path]:
    """
    Locate the MPI tester script. It is installed alongside e4s-cl and is used
    to load and run an MPI library for analysis
    """

    # First look in PATH for the script
    script = util.which(E4S_CL_MPI_TESTER_SCRIPT_NAME)

    if script is None:
        # If installed via pip, the script should be located in the same
        # directory as e4s-cl. Resolve argv[0] to bypass symlinks and look for
        # the file in the parent directory
        install_dir = Path(E4S_CL_SCRIPT).resolve().parent
        installed = Path(install_dir, E4S_CL_MPI_TESTER_SCRIPT_NAME)
        if installed.exists():
            script = installed

    if script is None:
        return None
    return Path(script)


def _generate_command(args: argparse.Namespace) -> List[str]:
    """
    Generate a command from the given args with a launcher and mpi binary
    """
    # Use the environment's MPI scripts by default
    launcher = util.which('mpirun')

    # If an installation dir is specified, use its contents
    if getattr(args, 'mpi', None):
        mpirun = Path(args.mpi) / "bin" / "mpirun"
        if mpirun.exists():
            launcher = mpirun.as_posix()

        # Update LD_LIBRARY_PATH if provided path exist
        mpi_lib_dir = Path(args.mpi) / "lib"
        if mpi_lib_dir.exists():
            util.prepend_library_path(mpi_lib_dir)

    # Use the launcher and passed as an argument in priority
    launcher = getattr(args, 'launcher', launcher)
    launcher_args = shlex.split(getattr(args, 'launcher_args', ''))

    if not launcher:
        LOGGER.error(
            "No MPI launcher detected. Please load an MPI module, use the "
            "`--mpi` or `--launcher` options to specify the launcher program "
            "to use.")
        return []

    # If no arguments were given, check the default behaviour of the launcher
    if not launcher_args:
        _check_mpirun(launcher)

    tester = _find_tester()
    if tester is None:
        raise ValueError("Cannot find tester script !")

    command = [
        launcher,
        *launcher_args,
        tester.as_posix(),
    ]
    LOGGER.info("Tracing MPI execution using:\n'%s'", " ".join(command))

    # Run the program using the detect command and get a file list
    return command


def _skip_analysis(args) -> bool:
    """Skip analysis step when certain conditions are met"""

    # If using shifter, do not try to profile a library
    # TODO Allow user to request full initialization
    if getattr(args, 'backend', '') == 'shifter':
        LOGGER.info("Skipping analysis to use shifter modules")
        return False

    return True


def _rename_profile(profile: Profile, requested_name: Optional[str]) -> None:
    """Rename the given profile with a given name or create a special unique
    hash for it"""
    current_name = profile.get('name')
    if current_name == INIT_TEMP_PROFILE_NAME:
        current_name = None

    if ((requested_name == current_name and requested_name is not None)
            or (current_name and requested_name is None)):
        # match (requested_name, current_name) with:
        #     (None, None) -> rename (create hash)
        #     (None, x) -> pass (keep x)
        #     (x, None) -> rename (set x)
        #     (x, y) -> rename (overwrite y with x)
        #     (x, x) -> pass

        # ==> Rename if !((r == c && r != None) || (c && !r))
        return True

    if requested_name is None or profile.name == INIT_TEMP_PROFILE_NAME:
        hash_ = util.hash256(json.dumps(profile))
        requested_name = f"default-{hash_[:16]}"

    try:
        controller = Profile.controller()

        # Erase any potential existing profile
        if controller.one({"name": requested_name}):
            controller.delete({"name": requested_name})

        controller.update({'name': requested_name}, profile.eid)
    except (UniqueAttributeError, ModelError) as err:
        LOGGER.error('Failed to rename profile: %s', err)
        return False

    return True


def _filter_files(
    files: Iterable[Path],
    blacklist: Iterable[Path],
    mpi_libraries: Iterable[Path],
) -> List[Path]:
    """
    Optimize the 'files' list by removing files contained in the blacklist
    or in the MPI installation directory, if one is identified
    """
    filtered = files

    # Remove files contained in the MPI installation tree
    mpi_install_dir = library_install_dir(mpi_libraries)
    if mpi_install_dir:
        filtered = filter(
            lambda file: not util.path_contains(mpi_install_dir, file),
            filtered,
        )

    # Remove files present in a blacklist.
    filtered = filter(
        lambda file: not file in blacklist,
        filtered,
    )

    # If a blacklist is used, remove files used by the interpreter that might
    # have slipped the blacklist definition
    if blacklist:
        filtered = filter(
            lambda file: 'conda' not in file.parts,
            filtered,
        )

    return list(filter(None, {*filtered, mpi_install_dir}))


def _filter_libraries(
    libraries: List[Path],
    blacklist: List[Path],
) -> List[Path]:
    """
    Remove libraries from the list passed as an argument. Removal depends on
    the contents of the blacklist or the file path that could indicate it is
    not relevant to the profile being constructed
    """
    filtered = libraries

    # Remove libraries present in a blacklist.
    filtered = filter(
        lambda file: not file in blacklist,
        filtered,
    )

    # If a blacklist is used, remove files used by the interpreter that might
    # have slipped the blacklist definition
    if blacklist:
        filtered = filter(
            lambda file: 'conda' not in file.parts,
            filtered,
        )

    return list(filtered)


def _set_defaults(args: argparse.Namespace, profile_eid: int):
    """
    Update the profile with the backend, image and source arguments from the command line
    """
    controller = Profile.controller()
    changes = {}

    for label in ["backend", "image", "source", "wi4mpi"]:
        if label in args:
            changes.update({label: getattr(args, label, "")})

    # Update the profile
    controller.update(
        changes,
        profile_eid,
    )


def _optimize_profile(args: argparse.Namespace, profile_eid: int) -> int:
    """
    After the analysis has succeeded, agglomerate bound files and rename profile
    """
    controller = Profile.controller()

    # File blacklist. This is calculated by running the detection script with
    # no MPI library, to list all the python files required when interpreting
    # the program.
    # When a command is passed to the init command, we do not use a python
    # script/interpreter and do not need to blacklist any files.
    blacklist = []
    tester_used = not bool(getattr(args, 'cmd', None))
    if tester_used:
        _, blacklist = opened_files([_find_tester(), "-n"])

    # Fetch the profile
    selected_profile = controller.one(profile_eid)

    # Get the profile data and cast to Path
    profile_files = list(map(Path, selected_profile.get('files', [])))
    profile_libraries = list(map(Path, selected_profile.get('libraries', [])))
    profile_mpi_libraries = filter_mpi_libs(profile_libraries)

    # Filter the files
    file_paths = _filter_files(
        profile_files,
        blacklist,
        profile_mpi_libraries,
    )
    file_strings = list(map(str, file_paths))
    file_strings.sort()

    # Filter the libraries
    library_paths = _filter_libraries(
        profile_libraries,
        blacklist,
    )
    library_strings = list(map(str, library_paths))
    library_strings.sort()

    # Update the profile
    controller.update(
        {
            'files': file_strings,
            'libraries': library_strings
        },
        selected_profile.eid,
    )

    _set_defaults(args, selected_profile.eid)

    requested_name = (getattr(args, 'profile_name', None)
                      or profile_mpi_name(profile_mpi_libraries))

    if requested_name == INIT_TEMP_PROFILE_NAME:
        LOGGER.error(
            "This name is reserved. Please use another name for your profile")
        return EXIT_FAILURE

    if not _rename_profile(selected_profile, requested_name):
        return EXIT_FAILURE
    return EXIT_SUCCESS


class InitCommand(AbstractCommand):
    """`init` macrocommand."""

    def _construct_parser(self) -> argparse.ArgumentParser:
        parser = get_parser(prog=self.command, description=self.summary)

        parser.add_argument(
            '--launcher',
            help="MPI launcher required to run a sample program.",
            metavar='launcher',
            type=binary_in_path,
            default=SUPPRESS,
            dest='launcher',
        )

        parser.add_argument(
            '--launcher_args',
            help="MPI launcher arguments required to run a sample program.",
            metavar='launcher_args',
            default=SUPPRESS,
            dest='launcher_args',
        )

        parser.add_argument(
            '--mpi',
            type=posix_path,
            help="Path of the MPI installation to use with this profile",
            default=SUPPRESS,
            metavar='/path/to/mpi',
        )

        parser.add_argument(
            '--source',
            help="Script to source before execution with this profile",
            metavar='script',
            default=SUPPRESS,
            dest='source',
        )

        parser.add_argument(
            '--image',
            help="Container image to use by default with this profile",
            metavar='/path/to/image',
            default=SUPPRESS,
            dest='image',
        )

        parser.add_argument(
            '--backend',
            help="Container backend to use by default with this profile."
            f" Available backends are: {', '.join(EXPOSED_BACKENDS)}",
            metavar='technology',
            default=SUPPRESS,
            dest='backend',
        )

        parser.add_argument(
            '--profile',
            help="Name of the profile to create. This will erase an existing"
            " profile with the same name !",
            metavar='profile_name',
            default=SUPPRESS,
            dest='profile_name',
        )

        parser.add_argument(
            '--wi4mpi',
            help=
            "Path to the Wi4MPI install directory to use with this profile",
            metavar='path',
            default=SUPPRESS,
            dest='wi4mpi',
        )

        parser.add_argument(
            'cmd',
            help="Custom command to analyze, instead of the default. Use this"
            " if your MPI implementation has specificities set at compile time",
            metavar='command',
            default=SUPPRESS,
            nargs=REMAINDER,
        )

        return parser

    def main(self, argv: List[str]) -> int:
        args = self._parse_args(argv)

        profile_data = {'name': INIT_TEMP_PROFILE_NAME}

        controller = Profile.controller()

        # Erase any leftover temporary profiles
        controller.delete({"name": profile_data['name']})

        # Create and select a profile for use
        profile_eid = controller.create(profile_data).eid
        controller.select(profile_eid)

        status = EXIT_SUCCESS
        if _skip_analysis(args):
            try:
                status = _analyze_binary(args)
            except KeyboardInterrupt:
                status = EXIT_FAILURE

        if status == EXIT_FAILURE:
            controller.delete(profile_eid)
            return status

        status = _optimize_profile(args, profile_eid)
        LOGGER.info("Created profile %s", controller.selected().get('name'))

        return status


SUMMARY = "Initialize %(prog)s with the accessible MPI library, and create a profile with the results."

COMMAND = InitCommand(__name__, summary_fmt=SUMMARY)



"""
E4S Storage Containers.

A storage container provides a record database, a persistent filesystem, and a key/value store.
The database is accessed via member methods like :any:`insert`, :any:`search`, etc.
The filesystem is accessed via its filesystem prefix (e.g. ``/usr/local/packages``) via :any:`prefix`.
The key/value store is accessed via the `[]` operator, i.e. treat the storage object like a dictionary.
"""

from abc import ABCMeta, abstractmethod
from e4s_cl.error import Error


class StorageError(Error):
    """Indicates a failure in the storage system."""

    message_fmt = ("%(value)s\n"
                   "%(hints)s\n")


class StorageRecord(dict):
    """A record in the storage container's database.
    
    Attributes:
        eid_type: Element identifier type.
        storage: Storage container whos database contains this record.
        eid: Element identifier value.
    """
    eid_type = str

    def __init__(self, storage, eid, element):
        super().__init__(element)
        self.storage = storage
        self.eid = eid

    def __hash__(self):
        return hash(self.eid)


class AbstractStorage():
    """Abstract base class for storage containers.
    
    A storage container provides a record database, a persistent filesystem, and a key/value store.
    The database is accessed via member methods like :any:`insert`, :any:`search`, etc.
    The filesystem is accessed via its filesystem prefix (e.g. ``/usr/local/packages``) via :any:`prefix`.
    The key/value store is accessed via the `[]` operator, i.e. treat the storage object like a dictionary.
    
    Attributes:
        name (str): The storage container's name, e.g. "system" or "user".
        prefix (str): Absolute path to the top-level directory of the container's filesystem.
        database (str): Database object implementing :any:`AbstractDatabase`. 
    """

    __metaclass__ = ABCMeta

    Record = StorageRecord

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    @abstractmethod
    def __len__(self):
        """Return the number of items in the key/value store."""

    @abstractmethod
    def __getitem__(self, key):
        """Retrieve a value from the key/value store."""

    @abstractmethod
    def __setitem__(self, key, value):
        """Store a value in the key/value store."""

    @abstractmethod
    def __delitem__(self, key):
        """Remove a value from the key/value store."""

    @abstractmethod
    def __contains__(self, key):
        """Returns True if ``key`` maps to a value is in the key/value store."""

    @abstractmethod
    def __iter__(self):
        """Iterate over keys in the key/value store."""

    @abstractmethod
    def iterkeys(self):
        """Iterate over keys in the key/value store."""

    @abstractmethod
    def itervalues(self):
        """Iterate over values in the key/value store."""

    @abstractmethod
    def iteritems(self):
        """Iterate over items in the key/value store."""

    @abstractmethod
    def connect_filesystem(self, *args, **kwargs):
        """Prepares the store filesystem for reading and writing."""

    @abstractmethod
    def disconnect_filesystem(self, *args, **kwargs):
        """Makes the store filesystem unreadable and unwritable."""

    @abstractmethod
    def connect_database(self, *args, **kwargs):
        """Open the database for reading and writing."""

    @abstractmethod
    def disconnect_database(self, *args, **kwargs):
        """Close the database for reading and writing."""

    @abstractmethod
    def prefix(self):
        """Get the filesystem prefix for file storage.
        
        The filesystem must be persistent and provide the usual POSIX filesystem calls.
        In particular, GNU software packages should be installable in the filesystem.
        
        Returns:
            str: Absolute path in the filesystem.
        """

    @abstractmethod
    def __enter__(self):
        """Initiates the database transaction."""

    @abstractmethod
    def __exit__(self, ex_type, value, traceback):
        """Finalizes the database transaction."""

    @abstractmethod
    def table(self, table_name):
        """Return a handle to a table.
        
        Return a handle to the named table or, if `table_name` is None, return the default table.
        
        Args:
            table_name (str): Name of the table or None.
            
        Returns:
            object: A database table object.
        """

    @abstractmethod
    def count(self, table_name=None):
        """Count the records in the database.
        
        Args:
            table_name (str): Name of the table to operate on.  See :any:`AbstractStorage.table`.
            
        Returns:
            int: Number of records in the table.
        """

    @abstractmethod
    def get(self, keys, table_name=None, match_any=False):
        """Find a single record.
        
        The behavior depends on the type of `keys`:
            * self.Record.eid_type: return the record with that element identifier.
            * dict: return the record with attributes matching `keys`.
            * list or tuple: return a list of records matching the elements of `keys`
            * None: return None.
        
        Args:
            keys: Fields or element identifiers to match.
            table_name (str): Name of the table to operate on.  See :any:`AbstractStorage.table`.
            match_any (bool): Only applies if `keys` is a dictionary.  If True then any key 
                              in `keys` may match or if False then all keys in `keys` must match.

        Returns:
            Record: The matching data record if `keys` was a self.Record.eid_type or dict.
            list: All matching data records if `keys` was a list or tuple.
            None: No record found or ``bool(keys) == False``.
            
        Raises:
            ValueError: Invalid value for `keys`.
        """

    @abstractmethod
    def search(self, keys=None, table_name=None, match_any=False):
        """Find multiple records.
        
        The behavior depends on the type of `keys`:
            * Record.eid_type: return the record with that element identifier.
            * dict: return all records with attributes matching `keys`.
            * list or tuple: return a list of records matching the elements of `keys`
            * None: return all records.
        
        Args:
            keys: Fields or element identifiers to match.
            table_name (str): Name of the table to operate on.  See :any:`AbstractStorage.table`.
            match_any (bool): Only applies if `keys` is a dictionary.  If True then any key 
                              in `keys` may match or if False then all keys in `keys` must match.

        Returns:
            list: Matching data records.
            
        Raises:
            ValueError: Invalid value for `keys`.
        """

    @abstractmethod
    def match(self, field, table_name=None, regex=None, test=None):
        """Find records where `field` matches `regex` or `test`.
        
        Either `regex` or `test` may be specified, not both.  
        If `regex` is given, then all records with `field` matching the regular expression are returned.
        If test is given then all records with `field` set to a value that caues `test` to return True are returned. 
        If neither is given, return all records where `field` is set to any value. 
        
        Args:
            table_name (str): Name of the table to operate on.  See :any:`AbstractStorage.table`.
            field (string): Name of the data field to match.
            regex (string): Regular expression string.
            test: Callable returning a boolean value.  

        Returns:
            list: Matching data records.
            
        Raises:
            ValueError: Invalid value for `keys`.
        """

    @abstractmethod
    def contains(self, keys, table_name=None, match_any=False):
        """Check if the specified table contains at least one matching record.
        
        The behavior depends on the type of `keys`:
            * self.Record.eid_type: check for the record with that element identifier.
            * dict: check for the record with attributes matching `keys`.
            * list or tuple: return the equivilent of ``map(contains, keys)``.
        
        Args:
            keys: Fields or element identifiers to match.
            table_name (str): Name of the table to operate on.  See :any:`AbstractStorage.table`.
            match_any (bool): Only applies if `keys` is a dictionary.  If True then any key 
                              in `keys` may match or if False then all keys in `keys` must match.

        Returns:
            bool: True if the table contains at least one matching record, False otherwise.
            
        Raises:
            ValueError: Invalid value for `keys`.
        """

    @abstractmethod
    def insert(self, data, table_name=None):
        """Create a new record.
        
        If the table doesn't exist it will be created.
        
        Args:
            data (dict): Data to insert in table.
            table_name (str): Name of the table to operate on.  See :any:`AbstractStorage.table`.
            
        Returns:
            Record: The new record.
        """

    @abstractmethod
    def update(self, fields, keys, table_name=None, match_any=False):
        """Update records.
        
        The behavior depends on the type of `keys`:
            * self.Record.eid_type: update the record with that element identifier.
            * dict: update all records with attributes matching `keys`.
            * list or tuple: apply update to all records matching the elements of `keys`.
        
        Args:
            fields (dict): Data to record.
            keys: Fields or element identifiers to match.
            table_name (str): Name of the table to operate on.  See :any:`AbstractStorage.table`.
            match_any (bool): Only applies if `keys` is a dictionary.  If True then any key 
                              in `keys` may match or if False then all keys in `keys` must match.
            
        Raises:
            ValueError: ``bool(keys) == False`` or invaild value for `keys`.
        """

    @abstractmethod
    def unset(self, fields, keys, table_name=None, match_any=False):
        """Update records by unsetting fields.
        
        Update only allows you to update a record by adding new fields or overwriting existing fields. 
        Use this method to remove a field from the record.
        
        The behavior depends on the type of `keys`:
            * self.Record.eid_type: update the record with that element identifier.
            * dict: update all records with attributes matching `keys`.
            * list or tuple: apply update to all records matching the elements of `keys`.
        
        Args:
            fields (list): Names of fields to remove from matching records.
            keys: Fields or element identifiers to match.
            table_name (str): Name of the table to operate on.  See :any:`AbstractStorage.table`.
            match_any (bool): Only applies if `keys` is a dictionary.  If True then any key 
                              in `keys` may match or if False then all keys in `keys` must match.
            
        Raises:
            ValueError: ``bool(keys) == False`` or invaild value for `keys`.
        """

    @abstractmethod
    def remove(self, keys, table_name=None, match_any=False):
        """Delete records.
        
        The behavior depends on the type of `keys`:
            * self.Record.eid_type: delete the record with that element identifier.
            * dict: delete all records with attributes matching `keys`.
            * list or tuple: delete all records matching the elements of `keys`.
        
        Args:
            keys: Fields or element identifiers to match.
            table_name (str): Name of the table to operate on.  See :any:`AbstractStorage.table`.
            match_any (bool): Only applies if `keys` is a dictionary.  If True then any key 
                              in `keys` may match or if False then all keys in `keys` must match.
            
        Raises:
            ValueError: ``bool(keys) == False`` or invaild value for `keys`.
        """

    @abstractmethod
    def purge(self, table_name=None):
        """Delete all records.

        Args:
            table_name (str): Name of the table to operate on.  See :any:`AbstractStorage.table`.
        """



from os import getcwd, environ, pathsep
import tests
from e4s_cl import config
from e4s_cl.cf.containers import (
    BackendUnsupported,
    BoundFile,
    Container,
    FileOptions,
)

CONFIG_EXECUTABLE = tests.ASSETS / 'bin' / 'apptainer-conf'
DEFAULT_CONFIGURATION = config.CONFIGURATION
TEST_CONFIGURATION = config.Configuration.create_from_string(f"""
backends:
  apptainer:
    executable: '{CONFIG_EXECUTABLE}'
    options: ['--nocolor', '-s']
    exec_options: ['--hostname', 'XxmycoolcontainerxX']
""")


class ContainerTestApptainer(tests.TestCase):

    def test_additional_options_config(self):
        container = Container(name='apptainer')
        command = ['']

        apptainer_command = container._prepare(command)
        for option in {'--nocolor', '-s', '--hostname', 'XxmycoolcontainerxX'}:
            self.assertNotIn(option, apptainer_command)

        config.update_configuration(TEST_CONFIGURATION)
        apptainer_command = container._prepare(command)
        self.assertContainsInOrder([
            '--nocolor',
            '-s',
            'exec',
            '--hostname',
            'XxmycoolcontainerxX',
        ], apptainer_command)

        config.update_configuration(DEFAULT_CONFIGURATION)
        apptainer_command = container._prepare(command)
        for option in {'--nocolor', '-s', '--hostname', 'XxmycoolcontainerxX'}:
            self.assertNotIn(option, apptainer_command)

    def test_additional_options_environment(self):
        container = Container(name='apptainer')
        command = ['']

        apptainer_command = container._prepare(command)
        for option in {'--nocolor', '-s', '--hostname', 'XxmycoolcontainerxX'}:
            self.assertNotIn(option, apptainer_command)

        environ['E4S_CL_APPTAINER_OPTIONS'] = "--nocolor -s"
        environ[
            'E4S_CL_APPTAINER_EXEC_OPTIONS'] = "--hostname XxmycoolcontainerxX"
        apptainer_command = container._prepare(command)
        self.assertContainsInOrder([
            '--nocolor',
            '-s',
            'exec',
            '--hostname',
            'XxmycoolcontainerxX',
        ], apptainer_command)

        del environ['E4S_CL_APPTAINER_OPTIONS']
        del environ['E4S_CL_APPTAINER_EXEC_OPTIONS']
        apptainer_command = container._prepare(command)
        for option in {'--nocolor', '-s', '--hostname', 'XxmycoolcontainerxX'}:
            self.assertNotIn(option, apptainer_command)

    def test_executable(self):
        """Assert the default apptainer executable comes from $PATH"""
        container = Container(name='apptainer')

        default_path = environ.get('PATH', '')
        environ['PATH'] = f"{tests.ASSETS / 'bin'}{pathsep}{default_path}"

        self.assertEqual(tests.ASSETS / 'bin' / 'apptainer',
                         container._executable())

        environ['PATH'] = default_path

    def test_executable_config(self):
        """Assert the apptainer executable is read from the configuration"""
        container = Container(name='apptainer')

        config.update_configuration(TEST_CONFIGURATION)
        self.assertEqual(tests.ASSETS / 'bin' / 'apptainer-conf',
                         container._executable())

        config.update_configuration(DEFAULT_CONFIGURATION)

    def test_executable_env(self):
        """Assert the apptainer executable is read from the environment"""
        container = Container(name='apptainer')

        environ['E4S_CL_APPTAINER_EXECUTABLE'] = str(tests.ASSETS / 'bin' /
                                                     'apptainer-env')
        self.assertEqual(tests.ASSETS / 'bin' / 'apptainer-env',
                         container._executable())

        del environ['E4S_CL_APPTAINER_EXECUTABLE']

    def test_executable_priority(self):
        """Assert the environment has precedence over config and config over default"""
        container = Container(name='apptainer')

        default_path = environ.get('PATH', '')
        environ['PATH'] = f"{tests.ASSETS / 'bin'}{pathsep}{default_path}"
        config.update_configuration(TEST_CONFIGURATION)
        environ['E4S_CL_APPTAINER_EXECUTABLE'] = str(tests.ASSETS / 'bin' /
                                                     'apptainer-env')

        self.assertEqual(tests.ASSETS / 'bin' / 'apptainer-env',
                         container._executable())

        del environ['E4S_CL_APPTAINER_EXECUTABLE']

        self.assertEqual(tests.ASSETS / 'bin' / 'apptainer-conf',
                         container._executable())

        config.update_configuration(DEFAULT_CONFIGURATION)

        self.assertEqual(tests.ASSETS / 'bin' / 'apptainer',
                         container._executable())

        environ['PATH'] = default_path



"""
Select the profile associated with the name passed as an argument.
The selected profile is internally recorded and will be used implicitly in several commands.
"""

from e4s_cl import EXIT_SUCCESS
from e4s_cl.cli import arguments
from e4s_cl.model.profile import Profile
from e4s_cl.cli.command import AbstractCommand


class ProfileSelectCommand(AbstractCommand):
    """``profile select`` subcommand."""
    def _construct_parser(self):
        usage = f"{self.command} <profile_name>"
        parser = arguments.get_model_identifier(Profile,
                                                prog=self.command,
                                                usage=usage,
                                                description=self.summary)
        return parser

    def main(self, argv):
        profile = self._parse_args(argv).profile
        Profile.controller().select(profile)
        return EXIT_SUCCESS


COMMAND = ProfileSelectCommand(
    __name__,
    summary_fmt=("Select a profile.\n"
                 "Use `profile list` to see all profiles."))



import tests
from e4s_cl.model.profile import Profile
from e4s_cl.cli.commands.profile.delete import COMMAND

PROFILE_NAMES = ['mpich', 'mvapich', 'openmpi']


class ProfileDeleteTest(tests.TestCase):

    def tearDown(self):
        Profile.controller().unselect()
        self.resetStorage()

    def test_delete(self):
        Profile.controller().create({'name': PROFILE_NAMES[0]})

        self.assertCommandReturnValue(0, COMMAND, [PROFILE_NAMES[0]])

        self.assertFalse(Profile.controller().one({'name': PROFILE_NAMES[0]}))

        Profile.controller().create({'name': PROFILE_NAMES[0]})
        Profile.controller().create({'name': PROFILE_NAMES[1]})
        Profile.controller().create({'name': PROFILE_NAMES[2]})

        self.assertCommandReturnValue(0, COMMAND, PROFILE_NAMES)

        self.assertFalse(Profile.controller().one({'name': PROFILE_NAMES[0]}))
        self.assertFalse(Profile.controller().one({'name': PROFILE_NAMES[1]}))
        self.assertFalse(Profile.controller().one({'name': PROFILE_NAMES[2]}))

    def test_existence(self):
        self.assertNotCommandReturnValue(0, COMMAND, ['test01'])

    def test_matching(self):
        Profile.controller().create({'name': PROFILE_NAMES[0]})
        Profile.controller().create({'name': PROFILE_NAMES[1]})
        Profile.controller().create({'name': PROFILE_NAMES[2]})

        # Fails due to 'm' alone not matching any profile
        self.assertNotCommandReturnValue(0, COMMAND, [PROFILE_NAMES[0][0]])
        
        # Succeeds due to '#a#' matching with a profile
        self.assertCommandReturnValue(0, COMMAND, ['#a#'])

        Profile.controller().create({'name': PROFILE_NAMES[1]})
        # Succeeds due to 'm#' matching with two profile
        self.assertCommandReturnValue(0, COMMAND, ['m#'])
        

        # Fails due to 'o' alone not matching any profile
        self.assertNotCommandReturnValue(0, COMMAND, [PROFILE_NAMES[2][0]])
        # Succeeds due to 'o#' matching a profile
        self.assertCommandReturnValue(0, COMMAND, ['o#'])

        self.assertFalse(Profile.controller().one({'name': PROFILE_NAMES[0]}))
        self.assertFalse(Profile.controller().one({'name': PROFILE_NAMES[1]}))
        self.assertFalse(Profile.controller().one({'name': PROFILE_NAMES[2]}))

    def test_wildcard(self):
        Profile.controller().create({'name': 'test.#'})
        Profile.controller().create({'name': 'test01'})
        Profile.controller().create({'name': 'test02'})
        Profile.controller().create({'name': 'test.1.1'})
        Profile.controller().create({'name': 'test.2.1'})

        self.assertCommandReturnValue(0, COMMAND, 'test0#')
        self.assertTrue(Profile.controller().one({'name': 'test.1.1'}))
        self.assertTrue(Profile.controller().one({'name': 'test.2.1'}))
        self.assertFalse(Profile.controller().one({'name': 'test01'}))
        self.assertFalse(Profile.controller().one({'name': 'test02'}))
        self.assertTrue(Profile.controller().one({'name': 'test.#'}))

        self.assertCommandReturnValue(0, COMMAND, 'test.#.1')
        self.assertFalse(Profile.controller().one({'name': 'test.1.1'}))
        self.assertFalse(Profile.controller().one({'name': 'test.2.1'}))
        self.assertTrue(Profile.controller().one({'name': 'test.#'}))

        self.assertCommandReturnValue(0, COMMAND, 'test.#')
        self.assertFalse(Profile.controller().one({'name': 'test.#'}))




"""
Module introducing docker backend support
Needs the docker module from extras
"""

import os
import sys
from typing import List
# pylint: disable=import-error
import docker
from e4s_cl.logger import get_logger
from e4s_cl.cf.containers import Container, FileOptions, BackendError, BackendNotAvailableError

LOGGER = get_logger(__name__)

NAME = 'docker'
EXECUTABLES = ['docker']
MIMES = []


class DockerContainer(Container):
    """
    Class used to abstract docker containers
    """

    def run(self, command: List[str], overload: bool = True) -> int:
        # Create the client from the environment
        client = docker.from_env()

        # Ensure the queried image is accessible
        try:
            image = client.images.get(self.image)
        except docker.errors.ImageNotFound as err:
            client.close()
            raise BackendError('docker') from err
        except docker.errors.APIError as err:
            client.close()
            raise BackendNotAvailableError('docker') from err

        mounts = []
        for file in self.bound:
            mounts.append(
                docker.types.Mount(
                    file.destination.as_posix(),
                    file.origin.as_posix(),
                    type='bind',
                    read_only=file.option == FileOptions.READ_ONLY))

        container_env = dict(os.environ)
        for key, val in self.env.items():
            if val is None:
                container_env.pop(key, None)
            else:
                container_env[key] = val

        try:
            outlog = client.containers.run(image,
                                           command,
                                           environment=container_env,
                                           stdout=True,
                                           stderr=True,
                                           mounts=mounts,
                                           remove=True)
        except docker.errors.ImageNotFound as err:
            raise BackendError('docker') from err
        except docker.errors.APIError as err:
            raise BackendNotAvailableError('docker') from err
        except docker.errors.ContainerError as err:
            LOGGER.error("Process in container %s failed with code %d:",
                         err.container.short_id, err.exit_status)
            for line in err.stderr.decode().split("\n"):
                LOGGER.error(line)
            return err.exit_status
        else:
            print(outlog.decode(), file=sys.stdout, end='')
        finally:
            client.close()

        return 0


CLASS = DockerContainer



"""TODO: FIXME: Docs

E4S Commander core software architecture.

E4S Commander follows the `Model-View-Controller (MVC)`_ architectural pattern. 
Packages in :py:mod:`e4s_cl.model` define models and controllers.  The `model` module
declares the model attributes as a dictionary named `ATTRIBUTES` in the form::

    <attribute>: {
        property: value,
        [[property: value], ...]
    }

Attribute properties can be anything you wish, including these special properties:

* ``type`` (str): The attribute is of the specified type.
* ``required`` (bool): The attribute must be specified in the data record. 
* ``model`` (str): The attribute associates this model with another data model.
* ``collection`` (str): The attribute associates this model with another data model.
* ``via`` (str): The attribute associates with another model "via" the specified attribute.

The `controller` module declares the controller as a subclass of :any:`Controller`.  Controller
class objects are not bound to any particular data record and so can be used to create, update,
and delete records.  Controller class instances **are** bound to a single data record.

Models may have **associations** and **references**.  An association creates a relationship between
an attribute of this model and an attribute of another (a.k.a foreign) model.  If the associated
attribute changes then the foreign model will update its associated attribute.  A reference indicates
that this model is being referened by one or more attributes of a foreign model.  If a record of the
referenced model changes then the referencing model, i.e. the foreign model, will update all referencing 
attributes. 

Associations, references, and attributes are constructed when :py:mod:`e4s_cl.model` is imported.

Examples:

    *One-sided relationship*
    ::

        class Pet(Controller):
            attributes = {'name': {'type': 'string'}, 'hungry': {'type': 'bool'}}
        
        class Person(Controller):
            attributes = {'name': {'type': 'string'}, 'pet': {'model': 'Pet'}}
            
    We've related a Person to a Pet but not a Pet to a Person.  That is, we can query a Person 
    to get information about their Pet, but we can't query a Pet to get information about their Person.  
    If a Pet record changes then the Person record that referenced the Pet record is notified of the 
    change. If a Person record changes then nothing happens to Pet.  The Pet and Person classes codify
    this relationship as:
    
        * Pet.references = set( (Person, 'pet') )
        * Pet.associations = {}
        * Person.references = set() 
        * Person.associations = {}
    
    Some example data for this model::
    
        {
         'Pet': {1: {'name': 'Fluffy', 'hungry': True}},
         'Person': {0: {'name': 'Bob', 'pet': 1}} 
        }
                      
    *One-to-one relationship*
    ::
    
        class Husband(Controller):
            attributes = {'name': {'type': 'string'}, 'wife': {'model': 'Wife'}}
            
        class Wife(Controller):
            attributes = {'name': {'type': 'string'}, 'husband': {'model': 'Husband'}}
    
    We've related a Husband to a Wife.  The Husband may have only one Wife and the Wife may have
    only one Husband.  We can query a Husband to get information about their Wife and we can query a
    Wife to get information about their Husband.  If a Husband/Wife record changes then the Wife/Husband 
    record that referenced the Husband/Wife record is notified of the change.  The Husband and Wife classes
    codify this relationship as:
    
        * Husband.references = set( (Wife, 'husband') )
        * Husband.associations = {}
        * Wife.references = set( (Husband, 'wife') )
        * Wife.associations = {}  
    
    Example data::
    
        {
         'Husband': {2: {'name': 'Filbert', 'wife': 1}},
         'Wife': {1: {'name': 'Matilda', 'husband': 2}}
        }
    
    *One-to-many relationship*
    ::
    
        class Brewery(Controller):
            attributes = {'address': {'type': 'string'}, 
                          'brews': {'collection': 'Beer', 
                                    'via': 'origin'}}
            
        class Beer(Controller):
            attributes = {'origin': {'model': 'Brewery'}, 
                          'color': {'type': 'string'}, 
                          'ibu': {'type': 'int'}}
            
    We've related one Brewery to many Beers.  A Beer has only one Brewery but a Brewery has zero or
    more Beers.  The `brews` attribute has a `via` property along with a `collection` property to specify 
    which attribute of Beer relates Beer to Brewery, i.e. you may want a model to have multiple 
    one-to-many relationship with another model.  The Brewery and Beer classes codify this relationship as:
    
        * Brewery.references = set()
        * Brewery.associations = {'brews': (Beer, 'origin')}
        * Beer.references = set()
        * Beer.associations = {'origin': (Brewery, 'brews')}
        
    Example data::
    
        {
         'Brewery': {100: {'address': '4615 Hollins Ferry Rd, Halethorpe, MD 21227',
                           'brews': [10, 12, 14]}},
         'Beer': {10: {'origin': 100, 'color': 'gold', 'ibu': 45},
                  12: {'origin': 100, 'color': 'dark', 'ibu': 15},
                  14: {'origin': 100, 'color': 'pale', 'ibu': 30}}
        }
    
    *Many-to-many relationship*
    ::
    
        class Actor(Controller):
            attributes = {'home_town': {'type': 'string'},
                          'appears_in': {'collection': 'Movie',
                                         'via': 'cast'}}
                                         
        class Movie(Controller):
            attributes = {'title': {'type': 'string'},
                          'cast': {'collection': 'Actor',
                                   'via': 'appears_in'}}
                                   
    An Actor may appear in many Movies and a Movie may have many Actors.  Changing the `cast` of a Movie
    notifies Actor and changing the movies an Actor `apppears_in` notifies Movie.  This relationship is
    codified as:
    
        * Actor.references = set()
        * Actor.associations = {'appears_in': (Movies, 'cast')}
        * Movie.references = set()
        * Movie.associations = {'cast': (Actor, 'appears_in')}
    
.. _Model-View-Controller (MVC): https://en.wikipedia.org/wiki/Model-view-controller
"""



import tests
from e4s_cl.model.profile import Profile
from e4s_cl.cli.commands.profile.copy import COMMAND


class ProfileCopyTest(tests.TestCase):
    def tearDown(self):
        Profile.controller().unselect()
        self.resetStorage()

    def test_copy(self):
        Profile.controller().create({'name': 'test01'})
        self.assertCommandReturnValue(0, COMMAND, ['test01', 'test02'])

        copy = Profile.controller().one({'name': 'test02'})
        self.assertTrue(copy)
        self.assertEqual(copy.get('name'), 'test02')

    def test_existence(self):
        self.assertNotCommandReturnValue(0, COMMAND, ['test01', 'test02'])



import re
import tests
from e4s_cl.cli.arguments import get_parser

parser_usage = 'do not use'
parser_description = 'This parser is a test'
parser_epilog = 'Bye bye'

optional_title = '--option'
optional_help = 'This is help'
optional_metavar = 'value'


class ArgumentsTests(tests.TestCase):
    def setUp(self):
        parser = get_parser(__name__,
                            usage=parser_usage,
                            description=parser_description,
                            epilog=parser_epilog)

        parser.add_argument(optional_title,
                            help=optional_help,
                            metavar=optional_metavar)

        self.help_string = parser.format_help()

    def test_optional(self):
        pattern = re.compile(
            f"{optional_title}.*{optional_metavar}.*{optional_help}")

        self.assertTrue(re.search(pattern, self.help_string))

    def test_usage(self):
        self.assertTrue(re.search(parser_usage, self.help_string))

    def test_description(self):
        self.assertTrue(re.search(parser_description, self.help_string))

    def test_epilog(self):
        self.assertTrue(re.search(parser_epilog, self.help_string))



import os
import sys
import traceback
from e4s_cl import EXIT_FAILURE, EXIT_WARNING, E4S_CL_SCRIPT
from e4s_cl import logger

LOGGER = logger.get_logger(__name__)


class Error(Exception):
    """Base class for all errors in E4S Container Launcher.
    
    Attributes:
        value: Some value attached to the error, typically a string but could be anything with a __str__ method.
        hints (list): String hints for the user to help resolve the error.
        show_backtrace (bool): Set to True to include a backtrace in the error message.
        message_fmt (str): Format string for the error message.
    """
    show_backtrace = False

    message_fmt = (
        "An unexpected %(typename)s exception was raised:\n"
        "\n"
        "%(value)s\n"
        "\n"
        "%(backtrace)s\n"
        "This is a bug in E4S Container Launcher.\n"
        "Please raise an issue on Github with the contents of '%(logfile)s'.")

    def __init__(self, value, *hints):
        """Initialize the Error instance.
        
        Args:
            value (str): Message describing the error.
            *hints: Hint messages to help the user resolve this error.
        """
        super(Error, self).__init__()
        self.value = value
        self.hints = list(hints)
        self.message_fields = {'logfile': logger.LOG_FILE}

    @property
    def message(self):
        fields = dict(self.message_fields, value=self.value)
        if not self.hints:
            hints_str = ''
        elif len(self.hints) == 1:
            hints_str = f"Hint: {self.hints[0]}\n"
        else:
            bullet_list = '\n  * '.join(self.hints)
            hints_str = f"Hints:\n  * {bullet_list}\n"
        fields['hints'] = hints_str
        return self.message_fmt % fields

    def handle(self, etype, value, tb):
        if self.show_backtrace:
            self.message_fields['backtrace'] = ''.join(
                traceback.format_exception(etype, value, tb)) + '\n'
        self.message_fields['typename'] = etype.__name__
        LOGGER.critical(self.message)
        return EXIT_FAILURE


class InternalError(Error):
    """Indicates that an internal error has occurred, i.e. a bug.
    
    These are bad and really shouldn't happen.
    """
    show_backtrace = True


class ConfigurationError(Error):
    """Indicates that E4S Container Launcher cannot succeed with the given parameters.
    
    This is most commonly caused by user error, e.g the user specifies measurement
    settings that are incompatible with the application.
    """
    message_fmt = (
        "%(value)s\n"
        "\n"
        "%(hints)s\n"
        "Cannot proceed with the given inputs.\n"
        "Please check the configuration for errors or raise an issue on Github."
    )

    def __init__(self, value, *hints):
        """Initialize the Error instance.
        
        Args:
            value (str): Message describing the error.
            *hints: Hint messages to help the user resolve this error.
        """
        if not hints:
            hints = [f"Try `{os.path.basename(E4S_CL_SCRIPT)} --help`"]
        super().__init__(value, *hints)

    def __str__(self):
        return self.value


class ModelError(InternalError):
    """Indicates an error in model data or the model itself."""

    def __init__(self, model, value):
        """Initialize the error instance.
        
        Args:
            model (Model): Data model.
            value (str): A message describing the error.  
        """
        super().__init__(f"{model.name}: {value}")
        self.model = model


class UniqueAttributeError(ModelError):
    """Indicates that duplicate values were given for a unique attribute."""

    def __init__(self, model, unique):
        """Initialize the error instance.
        
        Args:
            model (Model): Data model.
            unique (dict): Dictionary of unique attributes in the data model.  
        """
        super().__init__(model,
                         f"A record with one of '{unique}' already exists")


class ImmutableRecordError(ConfigurationError):
    """Indicates that a data record cannot be modified."""


class IncompatibleRecordError(ConfigurationError):
    """Indicates that a pair of data records are incompatible."""


class ProfileSelectionError(ConfigurationError):
    """Indicates an error while selecting a profile."""

    def __init__(self, value, *hints):
        from e4s_cl.cli.commands.profile.create import COMMAND as profile_create_cmd
        from e4s_cl.cli.commands.profile.select import COMMAND as profile_select_cmd
        from e4s_cl.cli.commands.profile.list import COMMAND as profile_list_cmd
        if not hints:
            hints = (
                f"Use `{profile_create_cmd}` to create a new profile configuration.",
                f"Use `{profile_select_cmd} <name>` to select a configuration.",
                f"Use `{profile_list_cmd}` to see available profile configurations."
            )
        super().__init__(value, *hints)


def excepthook(etype, value, tb):
    """Exception handler for any uncaught exception (except SystemExit).
    
    Replaces :any:`sys.excepthook`.
    
    Args:
        etype: Exception class.
        value: Exception instance.
        tb: Traceback object.
    """
    if etype == KeyboardInterrupt:
        LOGGER.info('Received keyboard interrupt. Exiting.')
        sys.exit(EXIT_WARNING)
    else:
        backtrace = ''.join(traceback.format_exception(etype, value, tb))
        LOGGER.debug(backtrace)
        if hasattr(value, 'handle'):
            sys.exit(value.handle(etype, value, tb))
        else:
            message = Error.message_fmt % {
                'value': value,
                'typename': etype.__name__,
                'logfile': logger.LOG_FILE,
                'backtrace': backtrace
            }
            LOGGER.critical(message)
            sys.exit(EXIT_FAILURE)


sys.excepthook = excepthook



import tests
from copy import deepcopy
from pathlib import Path
from e4s_cl.error import ProfileSelectionError
from e4s_cl.model.profile import Profile, homogenize_files

P_NAME = dict(name='profile')

BASE_DATA = dict(name='test',
                 files=['/dev/a', '/dev/b', '/dev/c'],
                 libraries=[])

PATCHES = [
    dict(name='new_name'),
    dict(files=['/etc/apptainer']),
    dict(libraries=['/not/a/lib.so'])
]


class ModelProfileTest(tests.TestCase):

    def tearDown(self):
        Profile.controller().unselect()
        self.resetStorage()
        self.assertFalse(Profile.controller().all())

    def test_create(self):
        controller = Profile.controller()
        profile = controller.create(BASE_DATA)

        created = controller.one(profile.eid)

        self.assertDictEqual(profile, BASE_DATA)
        self.assertDictEqual(created, BASE_DATA)

    def test_create_same_name(self):
        controller = Profile.controller()
        profile = controller.create(P_NAME)

        with self.assertRaises(Exception):
            controller.create(P_NAME)

    def test_create_no_name(self):
        controller = Profile.controller()

        with self.assertRaises(Exception):
            controller.create(dict(files=[]))

    def test_update_same_name(self):
        controller = Profile.controller()
        controller.create(P_NAME)
        profile = controller.create(dict(name='another_profile'))

        with self.assertRaises(Exception):
            controller.update(P_NAME, dict(name='another_profile'))

    def test_update_no_match(self):
        controller = Profile.controller()
        with self.assertRaises(Exception):
            controller.update(P_NAME, dict(name='another_profile'))

    def test_delete(self):
        controller = Profile.controller()

        controller.create(P_NAME)
        self.assertTrue(controller.all())
        controller.delete(P_NAME)
        self.assertFalse(controller.all())

        profile = controller.create(P_NAME)
        self.assertTrue(controller.all())
        controller.delete(profile)
        self.assertFalse(controller.all())

    def test_select_eid(self):
        controller = Profile.controller()

        profile = controller.create(P_NAME)
        controller.select(profile)
        self.assertEqual(controller.selected(), profile)

    def test_select_data(self):
        controller = Profile.controller()

        profile = controller.create(P_NAME)
        controller.select(P_NAME)
        self.assertEqual(controller.selected(), profile)

    def test_select_empty(self):
        controller = Profile.controller()

        with self.assertRaises(ProfileSelectionError):
            controller.selected()

    def test_select_no_match(self):
        controller = Profile.controller()
        with self.assertRaises(ProfileSelectionError):
            controller.select(P_NAME)

    def testHomogenization(self):
        paths = ['/proc/', '///mnt/../test///']
        posix = list(map(lambda x: Path(x).as_posix(), paths))
        data = dict(files=paths)
        homogenize_files(data)

        self.assertListEqual(posix, data['files'])

        data = dict()
        self.assertNotIn('files', data)


def wrapper_update(patch):

    def generated(self):
        controller = Profile.controller()
        profile = controller.create(BASE_DATA)

        expected = deepcopy(BASE_DATA)
        expected.update(patch)
        controller.update(patch, profile.eid)
        updated = controller.one(profile.eid)

        self.assertDictEqual(updated, expected)

    generated.__name__ = f"test_update_{' '.join(patch)}"
    return generated


for patch in PATCHES:
    test_update = wrapper_update(patch)
    setattr(ModelProfileTest, test_update.__name__, test_update)



"""Execute command

Definition of arguments and hooks related to the execute command,
file import calculations, and execution of a program passed as an
argument.
This command is used internally and thus cloaked from the UI
"""

import os
from typing import Union, List
from pathlib import Path
import re 
from sotools.linker import resolve
from sotools.libraryset import LibrarySet, Library
from e4s_cl import (
    E4S_CL_SCRIPT,
    EXIT_SUCCESS,
    config,
    logger,
    variables,
)
from e4s_cl.util import which
from e4s_cl.error import InternalError
from e4s_cl.cli import arguments
from e4s_cl.cli.command import AbstractCommand
from e4s_cl.cf.template import Entrypoint
from e4s_cl.cf.containers import Container, FileOptions
from e4s_cl.cf.libraries import (libc_version, library_links)
from e4s_cl.cf.wi4mpi import (
    wi4mpi_import,
    wi4mpi_libpath,
    wi4mpi_libraries,
    wi4mpi_preload,
    wi4mpi_root,
)

LOGGER = logger.get_logger(__name__)
_SCRIPT_CMD = Path(E4S_CL_SCRIPT).name


def import_library(shared_object, container):
    """
    This method binds the shared object it got as an argument, along with all
    the symbolic links that may exist and point to the same file.

    This is because depending on the linker at compile-time some binaries
    require more or less precise versions of the same file (eg. libmpi.so for
    some and libmpi.so.12 for others). Binding all the references ensures the
    library is found down the line.
    """
    for file in library_links(shared_object):
        container.bind_file(file, Path(container.import_library_dir,
                                       file.name))


# pylint: disable=unused-argument
def filter_libraries(library_set, container, entrypoint):
    """ Library filter

    library_paths: list[pathlib.Path]
    container: e4s_cl.cf.containers.Container

    This method filters out libraries that may cause problems when imported
    in a newer glibc environment
    """

    filtered_set = LibrarySet(library_set)

    # Remove libc-tied libraries from the filtered set, including the linker
    for lib in library_set.glib:
        filtered_set.remove(lib)

    return filtered_set


def overlay_libraries(library_set, container, entrypoint):
    """ Library overlay

    library_paths: list[pathlib.Path]
    container: e4s_cl.cf.containers.Container
    entrypoint: e4s_cl.cf.template.Entrypoint

    This method binds the libraries defined in the set, along with
    with the host's (implicitly newer) glib library suite and a bash binary
    to enable running scripts.

    Those bounds are necessary to run libraries that might depend on the more
    recent libraries of the host. glib binaries often use private symbols that
    are version dependent (GLIBC_PRIVATE) and this step ensures a match.
    """

    # Determine what the host's bash binary depends on
    full = LibrarySet.create_from([which('bash')])
    bash_binary = full.top_level
    bash_requirements = full - bash_binary

    # Hardcoded glib libraries sonames
    glib_sonames = [
        'libc.so.6',
        'libdl.so.2',
        'libcrypt.so.1',
        'libm.so.6',
        'libmvec.so.1',
        'libnsl.so.1',
        'libnss_compat.so.2',
        'libnss_db.so.2',
        'libnss_dns.so.2',
        'libnss_files.so.2',
        'libnss_hesiod.so.2',
        'libpthread.so.0',
        'libresolv.so.2',
        'librt.so.1',
    ]
    # Find the available libraries on the host, and bundle them in a set
    paths = filter(None, map(resolve, glib_sonames))
    glib_set = LibrarySet.create_from(paths)

    # Import the bash binary in the container
    entrypoint.interpreter = Path(container.import_binary_dir,
                                  'bash').as_posix()
    container.bind_file(bash_binary.pop().binary_path,
                        dest=entrypoint.interpreter)

    # Import the library set passed as an argument along with the bash
    # dependencies and the host's glib
    selected = LibrarySet(library_set | bash_requirements | glib_set)

    # Figure out what to if multiple linkers are required
    if len(selected.linkers) != 1:
        raise InternalError(
            f"{len(selected.linkers)} linkers detected. This should not happen."
        )

    # Override linkers in the container
    for linker in selected.linkers:
        entrypoint.linker = Path(container.import_binary_dir,
                                 Path(linker.binary_path).name).as_posix()
        container.bind_file(linker.binary_path, dest=entrypoint.linker)

    # Override the container's glib with the host's
    for lib in selected.glib | glib_set:
        if lib.soname in container.cache:
            LOGGER.debug("Overriding guest `%s` with host `%s`",
                         container.cache[lib.soname], lib.binary_path)
            container.bind_file(lib.binary_path,
                                dest=container.cache[lib.soname])

    # Remove all the glib libraries from the import list as they have been
    # bound above
    return LibrarySet(selected - glib_set)


def select_libraries(library_set, container, entrypoint):
    """ Select the libraries to make available in the future container

    This method checks the libc versions compatibilities and returns a
    list of libraries safe to bind dependending on that check.

    If the container's linker is more recent than the hosts, it must be used.
    libc.so's version must equals the linker's. Some libraries must be bound,
    some must not as they exist in the container, tied to its linker.
     => filter_libraries

    However, if the host linker is newer, the host libc is newer, and libraries
    fail if used with the older libc of the container. We must then bind it,
    but binding libc implies binding the linker too, as both need to match.
    In the process, we bind all the necessary libraries.
     => overlay_libraries

    Why not overlay in both cases you ask ? As we run container-compiled
    binaries, they expect a minimal version of libc. It is fine to run with
    a newer libc, but very hazardous to run with an older one.
    """

    host_newer = True
    guest_newer = False

    # Analyze the container to get glibc information
    container.get_data()
    guest_libc = container.libc_v
    host_libc = libc_version()

    methods = {host_newer: overlay_libraries, guest_newer: filter_libraries}

    host_precedence = host_libc > guest_libc

    LOGGER.debug("Host libc: %s %s Guest libc: %s", str(host_libc),
                 '>' if host_precedence else '<=', str(guest_libc))

    selected_libraries = methods[host_precedence](library_set, container,
                                                  entrypoint)

    for line in selected_libraries.ldd_format():
        LOGGER.debug(line)

    return selected_libraries


def generate_rtld_path(container: Container,
                       wi4mpi_install_dir: Path) -> List[Path]:
    """
    Create the final path list to be passed to LD_LIBRARY_PATH in the container
    """
    path_list = []

    if wi4mpi_install_dir is not None:
        wi4mpi_paths = list(
            map(lambda x: x.as_posix(), wi4mpi_libpath(wi4mpi_install_dir)))

        path_list += wi4mpi_paths

    if hasattr(container.__class__, 'linker_path'):
        path_list += container.__class__.linker_path

    return path_list + [container.import_library_dir]


def _check_access(path: Union[Path, str]) -> bool:
    check = Path(path).exists()

    if not check:
        LOGGER.debug("Omitting file '%s' from bind list: file not found",
                     Path(path).as_posix())

    return check


# ------------------- MPICH-family aliasing (no OMPI) -------------------

# Families we support for aliasing (MPICH lineage, including Cray)
# - C MPI:      libmpi.so.*, libmpi_cray.so.*
# - Fortran:    libmpifort.so.*, libmpifort_cray.so.*
# - C++ MPI:    libmpicxx.so.*, libmpi_cxx.so.*
#
# Explicitly excluded (Open MPI):
# - libmpi.so.4X (e.g., .40, .41, ...)
# - libmpi_{mpifh,usempi,usempif08}.so.*
_MPI_FAMILY_PATTERNS = {
    'mpi': re.compile(r'^libmpi(?:_cray)?\.so(?:\.(?!4[0-9])\d+)*$'),
    'mpifort': re.compile(r'^libmpifort(?:_cray)?\.so(?:\.\d+)*$'),
    'mpicxx': re.compile(r'^(?:libmpicxx|libmpi_cxx)\.so(?:\.\d+)*$'),
}
# Open MPI exclusions
_OMPI_MPI_RE = re.compile(r'^libmpi\.so\.4[0-9]+$')
_OMPI_FORTRAN_SPLIT_RE = re.compile(
    r'^libmpi_(?:mpifh|usempi|usempif08)\.so(?:\.\d+)*$')


def _which_mpi_family(name: str) -> str:
    for fam, pat in _MPI_FAMILY_PATTERNS.items():
        if pat.match(name):
            return fam
    return ''


def _lib_name_candidates(lib: Library) -> List[str]:
    """
    Return candidate names to classify a Library (prefer soname, fallback to filename).
    """
    names = []
    try:
        if lib.soname:
            names.append(lib.soname)
    except AttributeError:
        pass
    names.append(Path(lib.binary_path).name)
    return names


def alias_guest_mpi_sonames_conservative(library_set: LibrarySet,
                                         container: Container) -> None:
    """
    Bind additional alias filenames for MPICH-family MPI libs into the import dir,
    matching guest SONAMEs from the container's ld.so.cache. Excludes Open MPI.

    For each host library in a supported family (mpi, mpifort, mpicxx), and for
    each guest SONAME in the same family, bind:
      origin: <host lib path>
      dest:   /.e4s-cl/hostlibs/<guest SONAME>

    This ensures the guest loader resolves its expected SONAME to the host file.
    """
    if not container.cache:
        return

    # Build guest SONAMEs by family from ld.so.cache keys, excluding Open MPI
    guest_by_family = {fam: [] for fam in _MPI_FAMILY_PATTERNS.keys()}
    for soname in container.cache.keys():
        # Skip Open MPI variants explicitly
        if _OMPI_MPI_RE.match(soname) or _OMPI_FORTRAN_SPLIT_RE.match(soname):
            continue
        fam = _which_mpi_family(soname)
        if fam:
            guest_by_family[fam].append(soname)

    # For each host lib in the set, classify and alias to matching guest SONAMEs
    for lib in library_set:
        host_family = ''
        for candidate in _lib_name_candidates(lib):
            host_family = _which_mpi_family(candidate)
            if host_family:
                break
        if not host_family:
            continue  # not an MPI lib we manage

        guest_names = guest_by_family.get(host_family, [])
        if not guest_names:
            continue

        for gname in guest_names:
            dest = Path(container.import_library_dir, gname)
            LOGGER.debug("Aliasing host '%s' as guest '%s' at '%s'",
                         lib.binary_path, gname, dest)
            container.bind_file(lib.binary_path, dest=dest)


def _alias_wi4mpi_fake_libs_for_guest(wi4mpi_install_dir: Path,
                                      container: Container) -> None:
    """
    When using Wi4MPI, bind fake MPI libraries with guest SONAME aliases.
    
    Wi4MPI provides fake libraries (e.g., libmpi.so.12) that need to be accessible
    with MPICH-specific names (e.g., libmpich.so.12, libmpichfort.so.12) so that
    container binaries linked to MPICH can find them.
    
    This creates bindings so that guest MPI SONAMEs resolve to Wi4MPI fake libraries.
    """
    # Get the Wi4MPI FROM parameter to determine which fake library directory to use
    from_family = os.environ.get("WI4MPI_FROM", "").upper()
    
    if not from_family:
        LOGGER.debug("WI4MPI_FROM not set; skipping fake library aliasing")
        return
    
    LOGGER.debug("Wi4MPI fake lib aliasing: from_family=%s", from_family)
    fakelib_dir = wi4mpi_install_dir / 'libexec' / 'wi4mpi' / f"fakelib{from_family}"
    
    if not fakelib_dir.exists():
        LOGGER.warning("Wi4MPI fakelib directory not found: %s", fakelib_dir)
        return
    
    LOGGER.debug("Wi4MPI fake lib aliasing: fakelib_dir=%s", fakelib_dir)
    
    # Mapping from generic MPI library names to MPICH-specific names
    # Wi4MPI provides libmpi.so.*, libmpifort.so.*, etc.
    # MPICH uses libmpich.so.*, libmpichfort.so.*, etc.
    name_mappings = {
        'libmpi.so': ['libmpich.so', 'libmpi.so'],
        'libmpifort.so': ['libmpichfort.so', 'libmpifort.so'],
        'libmpicxx.so': ['libmpichcxx.so', 'libmpicxx.so'],
    }
    
    # Bind Wi4MPI fake libraries with appropriate SONAME aliases for MPICH
    # This ensures binaries linked to libmpich.so.X find the Wi4MPI translation layer
    for wi4mpi_base, target_bases in name_mappings.items():
        # Find all Wi4MPI fake libraries matching this base (e.g., libmpi.so.12, libmpi.so.4)
        fake_libs = list(fakelib_dir.glob(f"{wi4mpi_base}*"))
        if not fake_libs:
            continue
        
        for fake_lib in fake_libs:
            # Extract version suffix (e.g., ".12" from "libmpi.so.12")
            fake_name = fake_lib.name
            if '.' in fake_name and fake_name.count('.') >= 2:
                # Get suffix after base name (e.g., ".12" from "libmpi.so.12")
                suffix_start = fake_name.find('.', fake_name.find('.') + 1)
                suffix = fake_name[suffix_start:]
                
                # Create aliases for each target base name
                for target_base in target_bases:
                    target_soname = f"{target_base}{suffix}"
                    # Always create alias for MPICH variants when translating from MPICH
                    dest = Path(container.import_library_dir, target_soname)
                    LOGGER.debug("Aliasing Wi4MPI fake lib '%s' as guest '%s' at '%s'",
                               fake_lib, target_soname, dest)
                    container.bind_file(fake_lib, dest=dest)


class ExecuteCommand(AbstractCommand):
    """``execute`` subcommand."""

    def _construct_parser(self):
        usage = f"{self.command} [arguments] <command> [command_arguments]"
        parser = arguments.get_parser(prog=self.command,
                                      usage=usage,
                                      description=self.summary)
        parser.add_argument("--backend",
                            type=str,
                            dest='backend',
                            required=True,
                            help="Specify the backend name",
                            metavar='backend')

        parser.add_argument('--image',
                            type=str,
                            dest='image',
                            help="Container image to use",
                            metavar='image')

        parser.add_argument('--files',
                            type=arguments.posix_path_list,
                            help="Files to bind, comma-separated. Supports host:container syntax.",
                            default=[],
                            metavar='files')

        parser.add_argument('--libraries',
                            type=arguments.posix_path_list,
                            help="Libraries to bind, comma-separated",
                            default=[],
                            metavar='libraries')

        parser.add_argument('--source',
                            type=arguments.posix_path,
                            help="Script to source",
                            metavar='libraries')

        parser.add_argument('cmd',
                            type=str,
                            help="Executable command, e.g. './a.out'",
                            metavar='command',
                            nargs=arguments.REMAINDER)

        return parser

    def main(self, argv):
        args = self._parse_args(argv)

        # This object automatically mutates to one of the defined container
        # classes, and holds information about what is bound to the future
        # container launch
        container = Container(name=args.backend, image=args.image)

        # All execute does is build this object that translates to a bash
        # script, that is then bound to the container to be executed
        params = Entrypoint()

        # Check the environment for the use of Wi4MPI
        wi4mpi_install_dir = wi4mpi_root()

        # List the required libraries for proper Wi4MPI execution
        wi4mpi_required = []
        if wi4mpi_install_dir is not None:
            wi4mpi_required = wi4mpi_libraries(wi4mpi_install_dir)

        # The following is a set of all libraries required. It
        # is used in the container to check version mismatches
        required_libraries = [*args.libraries, *wi4mpi_required]

        # Filter out libraries that don't exist on this node. This can happen
        # in multi-node launches where filesystems differ across nodes (e.g.
        # ROCm libs present on some nodes but not others).
        available_libraries = []
        for lib_path in required_libraries:
            if Path(lib_path).exists():
                available_libraries.append(lib_path)
            else:
                LOGGER.warning("Library '%s' not found on this node, skipping",
                               lib_path)
        required_libraries = available_libraries
        
        # NOTE: Do NOT filter OpenMPI core libraries (libmpi.so.40, libopen-pal.so.40, etc.)
        # when Wi4MPI is active - Wi4MPI needs these to translate MPICH->OpenMPI calls.
        # Only the OpenMPI/PMIx plugin FILES should be filtered (see below).
        
        libset = LibrarySet.create_from(required_libraries)
        if libset:
            # Analyze the container to get library information from the environment
            # it offers, using the entrypoint and the above libraries
            container.get_data()

        # Bind all accessible requested files
        files = [p for p in (args.files or []) if Path(p) != Path('/')]
        if len(files) != len(args.files or []):
            LOGGER.warning("Filtered root '/' from bound files to avoid container permission errors.")

        for path in files:
            source, dest = path, None
            if ':' in path:
                # Split on rightmost colon to handle Unix paths that may contain colons
                parts = path.rsplit(':', 1)
                if len(parts) == 2:
                    source, dest = parts
                else:
                    LOGGER.error("Invalid file binding specification '%s'. Expected 'source:dest'.", path)
                    continue

            if dest is not None:
                if _check_access(source):
                    container.bind_file(source, dest=dest, option=FileOptions.READ_WRITE)
                else:
                    LOGGER.error("File '%s' not found.", source)
            else:
                if _check_access(path):
                    container.bind_file(path, option=FileOptions.READ_WRITE)
                else:
                    LOGGER.error("File '%s' not found.", path)

        # This script is sourced before any other command in the container
        params.source_script_path = args.source

        # Setup the final command and metadata relating to the execution
        params.command = args.cmd
        params.debug = logger.debug_mode()
        params.linker_library_path = generate_rtld_path(
            container, wi4mpi_install_dir)

        # Create a set of libraries to import using a library_set object,
        # then filtering it according to the contents of the container
        final_libset = select_libraries(libset, container, params)

        # Import each library along with all symlinks pointing to it
        for shared_object in final_libset:
            import_library(shared_object, container)

        # MPICH-family aliasing 
        # Skip aliasing when Wi4MPI is active, as it would bypass Wi4MPI translation
        if wi4mpi_install_dir is None:
            alias_guest_mpi_sonames_conservative(final_libset, container)
        else:
            # Wi4MPI is active: create aliases for Wi4MPI fake libraries to match guest SONAMEs
            # This ensures binaries linked to libmpich.so.12 find Wi4MPI's libmpi.so.12
            _alias_wi4mpi_fake_libs_for_guest(wi4mpi_install_dir, container)

        if wi4mpi_install_dir is None:
            # Preload the top-level libraries of the imported set to bypass
            # any potential RPATH settings
            def _path(library: Library):
                return Path(container.import_library_dir,
                            Path(library.binary_path).name).as_posix()

            if config.CONFIGURATION.preload_root_libraries:
                for imported_library_path in map(_path,
                                                 final_libset.top_level):
                    params.preload.append(imported_library_path)
        else:
            # If WI4MPI is found in the environment, import its files and
            # preload its required components
            wi4mpi_import(container, wi4mpi_install_dir)
            
            # Import target MPI libraries BEFORE constructing preload list
            # Wi4MPI uses WI4MPI_RUN_MPI_C_LIB to dlopen the target MPI library
            # The host path won't exist inside the container; we need to use
            # the container path where host libraries are bound
            host_mpi_lib = os.environ.get("WI4MPI_RUN_MPI_C_LIB", "")
            if host_mpi_lib:
                import_library(Library.from_path(Path(host_mpi_lib)), container)
                container_mpi_lib = Path(container.import_library_dir) / Path(host_mpi_lib).name
                params.extra_env["WI4MPI_RUN_MPI_C_LIB"] = container_mpi_lib.as_posix()
                LOGGER.debug("Wi4MPI: Overriding WI4MPI_RUN_MPI_C_LIB to container path: %s", container_mpi_lib)
            
            host_mpi_f_lib = os.environ.get("WI4MPI_RUN_MPI_F_LIB", "")
            if host_mpi_f_lib:
                import_library(Library.from_path(Path(host_mpi_f_lib)), container)
                container_mpi_f_lib = Path(container.import_library_dir) / Path(host_mpi_f_lib).name
                params.extra_env["WI4MPI_RUN_MPI_F_LIB"] = container_mpi_f_lib.as_posix()
                if host_mpi_lib:
                    params.extra_env["WI4MPI_RUN_MPIIO_C_LIB"] = container_mpi_lib.as_posix()
                params.extra_env["WI4MPI_RUN_MPIIO_F_LIB"] = container_mpi_f_lib.as_posix()
            
            # Now construct preload list AFTER libraries are imported
            # wi4mpi_preload will use container paths that we just ensured exist
            params.preload += wi4mpi_preload(wi4mpi_install_dir, container.import_library_dir)

        # Write the entry script to a file, then bind it to the container
        script_name = params.setup()
        container.bind_file(script_name, dest=container.script)

        command = [container.script]

        if variables.is_dry_run():
            LOGGER.info("Running %s in container %s", command, container)
            params.teardown()
            return EXIT_SUCCESS

        code = container.run(command)

        if code:
            LOGGER.critical("Container command failed with error code %d",
                            code)

        params.teardown()

        return code


COMMAND = ExecuteCommand(
    __name__,
    summary_fmt=
    "Execute a command in a container with a tailor-made environment.")



#!/bin/env python3

# Script that prints the given mpi's shared library version and distributor
# libmpi.so or equivalent is expected.

import os
import sys
from e4s_cl.cf.detect_name import _get_mpi_library_version

USAGE = "Usage: %s <libmpi.so.x>" % sys.argv[0]

DESCRIPTION = \
f"""{USAGE}

This program will extract and run the function associated to the 'MPI_Get_library_version' symbol from any shared object it is given.
"""

if __name__ == '__main__':
    if getattr(sys, 'frozen', False):
        __file__ = sys.executable

    if len(sys.argv) < 2:
        print(DESCRIPTION, end='')
        sys.exit(1)

    print(_get_mpi_library_version(sys.argv[1]), end='')
    sys.exit(0)



"""
CLI interface module
"""

import os
import re
import sys
from types import ModuleType
from e4s_cl import E4S_CL_SCRIPT, EXIT_FAILURE
from e4s_cl import logger, util
from e4s_cl.error import ConfigurationError, InternalError

LOGGER = logger.get_logger(__name__)

SCRIPT_COMMAND = os.path.basename(E4S_CL_SCRIPT)

COMMANDS_PACKAGE_NAME = __name__ + '.commands'

USAGE_FORMAT = "console"
"""Specify usage formatting:
    console: colorized and formatted to fit current console dimensions.
    markdown: plain text markdown. 
    completion: list available commands from the current cli.
"""

_COMMANDS = {SCRIPT_COMMAND: {}}


class UnknownCommandError(ConfigurationError):
    """Indicates that a specified command is unknown."""
    message_fmt = ("%(value)r is not a valid command.\n\n%(hints)s")


class AmbiguousCommandError(ConfigurationError):
    """Indicates that a specified partial command is ambiguous."""
    message_fmt = ("Command '%(value)s' is ambiguous.\n\n%(hints)s")

    def __init__(self, value, matches, *hints):
        parts = [
            f"Did you mean `{SCRIPT_COMMAND} {match}`?" for match in matches
        ]
        parts.append(f"Try `{SCRIPT_COMMAND} --help`")
        super().__init__(value, *hints + tuple(parts))


def _command_as_list(module_name):
    """Converts a module name to a command name list.
    
    Maps command module names to their command line equivilants, e.g.
    'e4s_cl.cli.commands.target.create' => ['e4s-cl', 'target', 'create']

    Args:
        module_name (str): Name of a module.

    Returns:
        list: Strings that identify the command.
    """
    parts = module_name.split('.')
    for part in COMMANDS_PACKAGE_NAME.split('.'):
        if parts[0] == part:
            parts = parts[1:]
    return [SCRIPT_COMMAND] + parts


def _get_commands(package_name):
    # pylint: disable=line-too-long
    """Returns a dictionary mapping commands to Python modules.
    
    Given a root module name, return a dictionary that maps commands and their
    subcommands to Python modules.  The special key ``__module__`` maps to the
    command module.  Other strings map to subcommands of the command.
    
    Args:
        package_name (str): A string naming the module to search for cli.
    
    Returns:
        dict: Strings mapping to dictionaries or modules.
        
    Example:
    ::

        _get_commands('e4s_cl.cli.commands.target') ==>
            {'__module__': <module 'e4s_cl.cli.commands.target' from '/home/jlinford/workspace/e4s_cl/packages/e4s_cl/cli/commands/target/__init__.pyc'>,
             'create': {'__module__': <module 'e4s_cl.cli.commands.target.create' from '/home/jlinford/workspace/e4s_cl/packages/e4s_cl/cli/commands/target/create.pyc'>},
             'delete': {'__module__': <module 'e4s_cl.cli.commands.target.delete' from '/home/jlinford/workspace/e4s_cl/packages/e4s_cl/cli/commands/target/delete.pyc'>},
             'edit': {'__module__': <module 'e4s_cl.cli.commands.target.edit' from '/home/jlinford/workspace/e4s_cl/packages/e4s_cl/cli/commands/target/edit.pyc'>},
             'list': {'__module__': <module 'e4s_cl.cli.commands.target.list' from '/home/jlinford/workspace/e4s_cl/packages/e4s_cl/cli/commands/target/list.pyc'>}}
    """
    def lookup(cmd, dct):
        if not cmd:
            return dct
        if len(cmd) == 1:
            return dct[cmd[0]]

        return lookup(cmd[1:], dct[cmd[0]])

    def walking_import(module, cmd, dct):
        car, cdr = cmd[0], cmd[1:]
        if cdr:
            walking_import(module, cdr, dct[car])
        elif car not in dct:
            __import__(module)
            dct.setdefault(car, {})['__module__'] = sys.modules[module]

    __import__(COMMANDS_PACKAGE_NAME)
    command_module = sys.modules[COMMANDS_PACKAGE_NAME]
    for _, module, _ in util.walk_packages(command_module.__path__,
                                           prefix=command_module.__name__ +
                                           '.'):
        if not (module.endswith('__main__') or '.tests' in module):
            try:
                lookup(_command_as_list(module), _COMMANDS)
            except KeyError:
                walking_import(module, _command_as_list(module), _COMMANDS)
    return lookup(_command_as_list(package_name), _COMMANDS)


def command_from_module_name(module_name):
    """Converts a module name to a command name string.
    
    Maps command module names to their command line equivilants, e.g.
    'e4s_cl.cli.commands.profile.show' => 'e4s-cl profile show'
    
    Args:
        module_name (str): Name of a module.
        
    Returns:
        str: A string that identifies the command.
    """
    if module_name == '__main__':
        return os.path.basename(E4S_CL_SCRIPT)

    return ' '.join(_command_as_list(module_name))


def commands_next(package_name=COMMANDS_PACKAGE_NAME):
    commands = sorted([
        i for i in _get_commands(package_name).items() if i[0] != '__module__'
    ])
    return [cmd for cmd, _ in commands]


def commands_description(package_name=COMMANDS_PACKAGE_NAME):
    """Builds listing of command names with short description.
    
    Args:
        package_name (str): A dot-seperated string naming the module to search for cli.
    
    Returns:
        str: Help string describing all commands found at or below `root`.
    """
    usage_fmt = USAGE_FORMAT.lower()
    groups = {}

    commands = sorted([
        i for i in _get_commands(package_name).items() if i[0] != '__module__'
    ])

    for cmd, topcmd in commands:
        module = topcmd['__module__']

        try:
            command_obj = module.COMMAND
        except AttributeError:
            continue

        if re.match(r'__.*', cmd):
            continue

        descr = command_obj.summary.split('\n')[0]
        group = command_obj.group

        if usage_fmt == 'console':
            command = util.color_text(f'{cmd:<14}', 'green')
            line = f"  {command}{descr}"
        elif usage_fmt == 'markdown':
            line = f"  {cmd:<28} | {descr}"
        else:
            line = ''

        groups.setdefault(group, []).append(line)

    parts = []

    for group, members in groups.items():
        title = group.title() + ' Subcommands' if group else 'Subcommands'

        if usage_fmt == 'console':
            parts.append(util.color_text(title + ':', attrs=['bold']))

        elif usage_fmt == 'markdown':
            parts.extend([
                '', ' ', f'{title:<30}' + ' | Description',
                f"{'-' * 30}:| {'-' * len('Description')}"
            ])

        parts.extend(members)
        parts.append('')

    return '\n'.join(parts)


def get_all_commands(package_name=COMMANDS_PACKAGE_NAME):
    """Builds a list of all commands and subcommands.

    Args:
        package_name (str): A dot-separated string naming the module to search for cli.

    Returns:
        list: List of modules corresponding to all commands and subcommands.
    """
    all_commands = []
    commands = sorted((i for i in _get_commands(package_name).items()
                       if i[0] != '__module__'))
    for _, topcmd in commands:
        for _, mod in topcmd.items():
            if isinstance(mod, dict):
                all_commands.append(mod['__module__'].__name__)
            elif isinstance(mod, ModuleType):
                all_commands.append(mod.__name__)
            else:
                raise InternalError(f"{mod} is an invalid module.")
    return all_commands


# pylint: disable=inconsistent-return-statements
def _resolve(cmd, c, d):
    # pylint: disable=invalid-name
    if not c:
        return []
    car, cdr = c[0], c[1:]
    try:
        matches = [(car, d[car])]
    except KeyError:
        matches = [i for i in d.items() if i[0].startswith(car)]
    if len(matches) == 1:
        return [matches[0][0]] + _resolve(cmd, cdr, matches[0][1])
    if len(matches) == 0:
        raise UnknownCommandError(' '.join(cmd))
    if len(matches) > 1:
        raise AmbiguousCommandError(' '.join(cmd), [m[0] for m in matches])


def find_command(cmd):
    """Import the command module and return its COMMAND member.
    
    Args:
        cmd (list): List of strings identifying the command, i.e. from :any:`_command_as_list`.
        
    Raises:
        UnknownCommandError: `cmd` is invalid.
        AmbiguousCommandError: `cmd` is ambiguous.
        
    Returns:
        AbstractCommand: Command object for the subcommand.
    """
    if cmd:
        root = '.'.join([COMMANDS_PACKAGE_NAME] + cmd)
    else:
        root = COMMANDS_PACKAGE_NAME
    try:
        return _get_commands(root)['__module__'].COMMAND
    except KeyError:
        LOGGER.debug('%r not recognized as a command', cmd)
        resolved = _resolve(cmd, cmd, _COMMANDS[SCRIPT_COMMAND])
        LOGGER.debug('Resolved ambiguous command %r to %r', cmd, resolved)
        return find_command(resolved)
    except AttributeError as a_err:
        raise InternalError(f"'COMMAND' undefined in {cmd}") from a_err


def _permute(cmd, cmd_args):
    cmd_len = len(cmd)
    full_len = len(cmd) + len(cmd_args)
    skip = [x[0] == '-' or os.path.isfile(x) for x in cmd + cmd_args]
    yield cmd, cmd_args
    for i in range(full_len):
        if skip[i]:
            continue
        for j in range(i + 1, full_len):
            if skip[j]:
                continue
            perm = cmd + cmd_args
            perm[i], perm[j] = perm[j], perm[i]
            yield perm[:cmd_len], perm[cmd_len:]


def execute_command(cmd, cmd_args=None, parent_module=None):
    """Import the command module and run its main routine.
    
    Partial commands are allowed, e.g. cmd=['e4s-cl', 'cli', 'commands', 'app', 'cre'] will resolve
    to 'e4s_cl.cli.commands.application.create'.  If the command can't be found then the parent 
    command (if any) will be invoked with the ``--help`` flag.
    
    Args:
        cmd (list): List of strings identifying the command, i.e. from :any:`_command_as_list`.
        cmd_args (list): Command line arguments to be parsed by command.
        parent_module (str): Dot-seperated name of the command's parent.
        
    Raises:
        UnknownCommandError: `cmd` is invalid.
        AmbiguousCommandError: `cmd` is ambiguous.
        
    Returns:
        int: Command return code.
    """
    if cmd_args is None:
        cmd_args = []

    if parent_module:
        parent = _command_as_list(parent_module)[1:]
        cmd = parent + cmd

    return find_command(cmd).main(cmd_args)



"""
Print a list of all the recorded profiles matching a given pattern, along with a brief description.

Pass option :code:`-s` to print only the names and disable formatting.

.. code::

    $ e4s-cl profile list 
    == Profile Configurations (/home/user/.local/e4s_cl/user.json) ===========

    +----------+---------------+---------+-------+-----------+-------+
    | Selected |     Name      | Backend | Image | Libraries | Files |
    +==========+===============+=========+=======+===========+=======+
    |          |   MPICH@3.4.2 |     N/A |   N/A |    16     |   1   |
    +----------+---------------+---------+-------+-----------+-------+
    |    *     | OpenMPI@4.1.1 |     N/A |   N/A |    21     |  73   |
    +----------+---------------+---------+-------+-----------+-------+

    $ e4s-cl profile list MPIC
    == Profile Configurations (/home/user/.local/e4s_cl/user.json) =========

    +----------+-------------+---------+-------+-----------+-------+
    | Selected |    Name     | Backend | Image | Libraries | Files |
    +==========+=============+=========+=======+===========+=======+
    |          | MPICH@3.4.2 |     N/A |   N/A |    16     |   1   |
    +----------+-------------+---------+-------+-----------+-------+
    

    $ e4s-cl profile list -s
    MPICH_3.4.2
    OpenMPI_4.1.1


Configuration
++++++++++++++

The following configuration option is available:

.. list-table::
   :widths: 10 5 5 20
   :header-rows: 1

   * - Configuration variable
     - Type
     - Default
     - Description

   * - :code:`profile_list_columns`
     - list
     - :code:`["selected", "name", "backend", "image"]`
     - Columns to display when running the `profile list` command. Available columns are: :code:`selected`, :code:`name`, :code:`libraries`, :code:`files`, :code:`backend` and :code:`image`.
"""

from typing import List, Dict, Callable
from e4s_cl import PROFILE_LIST_DEFAULT_COLUMNS
from e4s_cl.logger import get_logger
from e4s_cl.cli.cli_view import ListCommand
from e4s_cl.model.profile import Profile
from e4s_cl import config

LOGGER = get_logger(__name__)


def _count(attr: str) -> Callable[[Dict], int]:
    """
    Display a count of the attribute's items
    """
    return lambda x: len(x.get(attr, []))


def _selected(attr: str) -> Callable[[Dict], str]:
    """
    Display an asterisk if the given profile data matches the selected profile
    """
    return lambda x: '*' if Profile.selected().get(attr) == x[attr] else ' '


# All available columns and the actions they require
DEFINED_DASHBOARD_COLUMNS = [{
    'header': 'Selected',
    'function': _selected('name')
}, {
    'header': 'Name',
    'value': 'name',
    'align': 'r'
}, {
    'header': 'Backend',
    'value': 'backend',
    'align': 'r'
}, {
    'header': 'Image',
    'value': 'image',
    'align': 'r'
}, {
    'header': 'Libraries',
    'function': _count('libraries')
}, {
    'header': 'Files',
    'function': _count('files')
}]


def _valid_columns(names: List[str]) -> List[Dict]:
    """
    Return a list of column definitions as requested by the entry list of column names
    """

    def column(name):
        """
        Compare lowercase names to support case insensitivity
        """
        matches = list(
            filter(lambda x: x['header'].lower() == name.lower(),
                   DEFINED_DASHBOARD_COLUMNS))
        if matches:
            return matches[0]

        LOGGER.warning("Configuration error: Unrecognized column name: %s",
                       name)
        return None

    # Return a list of all the columns, filtering out Nones
    return list(filter(None, map(column, names)))


class ProfileListCommand(ListCommand):
    """
    Abstraction of the ListCommand to define profile specific fields to
    show in the list.
    """

    def __init__(self):
        selected_columns = _valid_columns(
            config.CONFIGURATION.profile_list_columns)
        if not selected_columns:
            selected_columns = _valid_columns(PROFILE_LIST_DEFAULT_COLUMNS)

        super().__init__(Profile, __name__, dashboard_columns=selected_columns)


COMMAND = ProfileListCommand()



"""
The :code:`launch` command is responsible for executing commands in containers with the requested MPI environment.

Running a command is done by prefixing a regular MPI launcher command with :code:`e4s-cl launch`.

Using profiles
^^^^^^^^^^^^^^

The preferred way of passing options is by using a \
:ref:`selected profile<profile_select>` or profile specified with the \
:code:`--profile` option. This way, the contents of the given profile are \
implicitly used, bypassing the need to specify any of the options manually. \
However, options given on the command line options have precedence over \
profiles' fields.

MPI translation
^^^^^^^^^^^^^^^

In case the MPI library used on the host and container do not belong to the \
same family, the :code:`--from` option can be used to specify which MPI family \
the binary was compiled with. **e4s-cl** will then take measures to translate \
MPI calls from the binary's MPI to the one passed as an argument. This is done \
using `Wi4MPI <https://github.com/cea-hpc/wi4mpi>`_.
"""

import os
import shlex
from pathlib import Path
from argparse import Namespace
from typing import Tuple, List
from dataclasses import (dataclass, field)
from e4s_cl import (
    E4S_CL_SCRIPT,
    EXIT_SUCCESS,
    config,
    logger,
    variables,
)
from e4s_cl.cli import arguments
from e4s_cl.util import run_e4scl_subprocess
from e4s_cl.cli.command import AbstractCommand
from e4s_cl.cf.launchers import interpret, get_reserved_directories
from e4s_cl.model.profile import Profile
from e4s_cl.cf.containers import EXPOSED_BACKENDS
from e4s_cl.cf.detect_mpi import (
    detect_mpi,
    filter_mpi_libs,
    library_install_dir,
)
from e4s_cl.cf.wi4mpi import (
    MPIFamily,
    SUPPORTED_TRANSLATIONS,
    WI4MPI_ENVIRONMENT_VARIABLES,
    WI4MPI_SOURCES,
    wi4mpi_find_libraries,
    wi4mpi_get_metadata,
    wi4mpi_prepare_environment_interface,
    wi4mpi_prepare_environment_preload,
)
from e4s_cl.cf.wi4mpi.install import install_wi4mpi

from e4s_cl.cli.commands.__execute import COMMAND as EXECUTE_COMMAND

LOGGER = logger.get_logger(__name__)
_SCRIPT_CMD = os.path.basename(E4S_CL_SCRIPT)


@dataclass
class Parameters:
    image: str = None
    backend: str = None
    source: Path = None
    libraries: set = field(default_factory=set)
    files: set = field(default_factory=set)
    wi4mpi: Path = None


def _parameters(args: dict) -> Parameters:
    """Generate compound parameters by merging profile and cli arguments
    The profile's parameters have less priority than the ones specified on
    the command line."""
    if isinstance(args, Namespace):
        args = vars(args)

    profile_data = dict(args.get('profile', {}))

    params = Parameters()

    for attribute, factory in [
        ('image', str),
        ('backend', str),
        ('libraries', lambda x: set(map(Path, x))),
        ('files', lambda x: set(map(Path, x))),
        ('source', lambda x: Path(x) if x else None),
        ('wi4mpi', lambda x: Path(x) if x else None),
    ]:
        value = args.get(attribute) or profile_data.get(attribute)
        if value is not None:
            setattr(params, attribute, factory(value))

    return params


def _format_execute(parameters: Parameters) -> List[str]:
    execute_command = shlex.split(str(EXECUTE_COMMAND))
    execute_command = [E4S_CL_SCRIPT] + execute_command[1:]

    if logger.debug_mode():
        execute_command = [execute_command[0], '-v'] + execute_command[1:]

    for attr in ['image', 'backend', 'source']:
        value = getattr(parameters, attr, None)
        if value:
            execute_command += [f"--{attr}", str(value)]

    for attr in ['libraries', 'files']:
        value = getattr(parameters, attr, None)
        if value:
            execute_command += [f"--{attr}", ",".join(map(str, value))]

    return execute_command


def _setup_wi4mpi(
    parameters: Parameters,
    translation: Tuple,
    family_metadata: MPIFamily,
    mpi_libraries: List[Path],
) -> List[str]:
    """Prepare the environment for the use of Wi4MPI
    - Set or update 'wi4mpi' in the parameters
    - Prepare the wi4mpi launcher for the final command
    """

    LOGGER.debug("Setting up Wi4MPI to translate '%s' to '%s' (%s)",
                 *translation, family_metadata.vendor_name)

    # Switch translation from mvapich to mpich, as wi4mpi doesn't recognise
    # mvapich as a separate flag
    if 'mvapich' in translation:
        translation = [vendor.replace('mvapich', 'mpich') for vendor in translation]

    # Locate the Wi4MPI installation and store it in parameters
    target_dir = parameters.wi4mpi
    if target_dir is None:
        target_dir = Path(config.CONFIGURATION.wi4mpi_install_directory)

    LOGGER.debug("Target: %s", target_dir)
    wi4mpi_install = install_wi4mpi(target_dir)
    if wi4mpi_install:
        parameters.wi4mpi = wi4mpi_install
    else:
        LOGGER.error(
            "Wi4MPI is required for this configuration, but installation failed"
        )
        return []

    run_c_lib, run_f_lib = wi4mpi_find_libraries(family_metadata,
                                                 mpi_libraries)

    if not (run_c_lib and run_f_lib):
        LOGGER.error(
            "Could not determine MPI libraries to use; Wi4MPI use aborted "
            "(no %(c_soname)s or %(f_soname)s in %(list)s)",
            {
                "c_soname": family_metadata.mpi_c_soname,
                "f_soname": family_metadata.mpi_f_soname,
                "list": list(mpi_libraries),
            },
        )
        return []

    # Deduce the MPI installation directory, add it to the imported files
    mpi_install_dir = library_install_dir([run_c_lib, run_f_lib])
    if mpi_install_dir:
        if mpi_install_dir == Path('/'):
            LOGGER.debug("Refusing to bind '/' as MPI install dir; binding library files instead.")
            # If the install dir is root (e.g. system install), we cannot bind root.
            # Bind the specific library files instead of the directory to avoid creating
            # an overlay that hides container system libraries (like libmpich.so).
            if run_c_lib:
                parameters.files.add(run_c_lib)
                # Helper: bind openmpi plugins if present (common in system installs)
                # Typically in <libdir>/openmpi
                plugin_dir = run_c_lib.parent / 'openmpi'
                if plugin_dir.exists():
                    parameters.files.add(plugin_dir)

                # Helper: bind pmix plugins if present (external pmix in system installs)
                for pmix_name in ['pmix', 'pmix2']:
                    pmix_dir = run_c_lib.parent / pmix_name
                    if pmix_dir.exists():
                        parameters.files.add(pmix_dir)

            if run_f_lib:
                parameters.files.add(run_f_lib)

            # Helper: bind /usr/share/openmpi for help files/config
            share_dir = Path('/usr/share/openmpi')
            if share_dir.exists():
                parameters.files.add(share_dir)

        else:
            parameters.files.add(mpi_install_dir)

    wi4mpi_wrapper = parameters.wi4mpi / 'bin' / 'wi4mpi'

    # Prepare the environment; add the Wi4MPI --from --to options
    if translation[0] == 'interface':
        wi4mpi_prepare_environment_interface(
            parameters.wi4mpi,
            family_metadata,
            mpi_install_dir,
            run_c_lib,
            run_f_lib,
        )
        wi4mpi_call = shlex.split(f"{wi4mpi_wrapper} -t {translation[1]}")
    else:
        wi4mpi_prepare_environment_preload(
            parameters.wi4mpi,
            translation[0],
            family_metadata,
            mpi_install_dir,
            run_c_lib,
            run_f_lib,
        )
        wi4mpi_call = shlex.split(
            f"{wi4mpi_wrapper} -f {translation[0]} -t {translation[1]}")

    return wi4mpi_call


class LaunchCommand(AbstractCommand):
    """``launch`` subcommand."""

    def _construct_parser(self):
        usage = f"{self.command} [arguments] [launcher] [launcher_arguments] [--] <command> [command_arguments]"
        parser = arguments.get_parser(
            prog=self.command,
            usage=usage,
            description=self.summary,
        )

        parser.add_argument(
            '--profile',
            type=arguments.single_defined_object(Profile, 'name'),
            help=
            "Profile to use. Its fields will be used by default, but any other argument will override them",
            default=Profile.selected().get('name', arguments.SUPPRESS),
            metavar='profile',
        )

        parser.add_argument(
            '--image',
            type=str,
            help="Path to the container image to run the program in",
            metavar='image',
        )

        parser.add_argument(
            '--backend',
            help="Container backend to use to launch the image." +
            f" Available backends are: {', '.join(EXPOSED_BACKENDS)}",
            metavar='technology',
            dest='backend',
        )

        parser.add_argument(
            '--source',
            type=arguments.posix_path,
            help="Path to a bash script to source before execution",
            metavar='source',
        )

        parser.add_argument(
            '--files',
            type=arguments.posix_path_list,
            help="Comma-separated list of files to bind",
            metavar='files',
        )

        parser.add_argument(
            '--libraries',
            type=arguments.posix_path_list,
            help="Comma-separated list of libraries to bind",
            metavar='libraries',
        )

        parser.add_argument(
            '--wi4mpi',
            type=arguments.posix_path,
            help="Path to a Wi4MPI installation (will be installed there if missing)",
            metavar='installation',
        )

        parser.add_argument(
            '--from',
            type=str.lower,
            choices=WI4MPI_SOURCES,
            help=
            "MPI family the command was intended to be run. Use this argument "
            "to toggle MPI call translation. Available families: " +
            ", ".join(WI4MPI_SOURCES),
            default=arguments.SUPPRESS,
            metavar='mpi-family',
        )

        parser.add_argument(
            'cmd',
            help="Executable command, e.g. './a.out'",
            metavar='command',
            nargs=arguments.REMAINDER,
        )

        return parser

    def main(self, argv):
        args = self._parse_args(argv)

        if getattr(args, 'profile', None) and '--profile' not in argv:
            LOGGER.info("Using selected profile %s", args.profile.get('name'))

        if not args.cmd:
            self.parser.error("No executable was specified.")

        # Merge profile and cli arguments to get a definitive list of arguments
        parameters = _parameters(args)

        if parameters.files:
            root_paths = {p for p in parameters.files if str(p) == "/"}
            if root_paths:
                parameters.files.difference_update(root_paths)
                LOGGER.warning(
                    "Filtered root '/' from bound files to avoid container permission errors."
                )

        # Ensure the minimum fields required for launch are present
        for _field in ['backend', 'image']:
            if not getattr(parameters, _field, None):
                if getattr(parameters, 'backend',None) == 'barebones':
                    continue
                self.parser.error(
                    f"Missing field: '{_field}'. Specify it using the "
                    "appropriate option or by selecting a profile.")

        launcher, program = interpret(args.cmd)

        for path in get_reserved_directories(launcher):
            if path.exists() and path not in parameters.files:
                parameters.files.add(path)

        wi4mpi_call = []

        # Source MPI family from the command line
        binary_mpi_family = getattr(args, 'from', None)
        # MPI shared objects from the parameters
        profile_mpi_libraries = filter_mpi_libs(parameters.libraries)
        # MPI family from the parameters (MPIIdentifier)
        profile_mpi_family = detect_mpi(profile_mpi_libraries)
        # MPI family Wi4MPI data from the parameters (MPIFamily)
        profile_mpi_family_data = wi4mpi_get_metadata(profile_mpi_family)

        if profile_mpi_family:
            LOGGER.debug(
                "Parameters contain the MPI library '%s'",
                profile_mpi_family,
            )
        else:
            LOGGER.debug(
                "No MPI family could be identified from the parameters")

        # If the parameters' MPI family is understood by Wi4MPI
        if profile_mpi_family_data:
            # Check the translation is supported
            translation = (binary_mpi_family, profile_mpi_family_data.cli_name)
            if translation in SUPPORTED_TRANSLATIONS:
                # Setup the environment and return the command line wrapper
                wi4mpi_call = _setup_wi4mpi(
                    parameters,
                    translation,
                    profile_mpi_family_data,
                    profile_mpi_libraries,
                )

                # Explicitly pass the environment if OpenMPI's launcher is used
                if (profile_mpi_family.vendor == "Open MPI"
                        and Path(launcher[0]).name == "mpirun"):
                    for varname in WI4MPI_ENVIRONMENT_VARIABLES:
                        if varname in os.environ.keys():
                            launcher.extend(shlex.split(f"-x {varname}"))

        # Helper: bind system configuration for libraries that use plugins/drivers
        # This prevents the host library from loading incompatible container drivers
        # by masking the container's configuration directories with the host's.
        
        # Define known configuration paths for libraries that require masking
        system_config_dirs = ['/etc', '/usr/etc', '/usr/local/etc']
        
        # Library -> Subdirectories to mask
        config_masks = {
            'libibverbs.so': ['libibverbs.d'],
            'libOpenCL.so': ['OpenCL'],
            'libvulkan.so': ['vulkan'],
            'libGLX.so': ['glvnd'],
            'libEGL.so': ['glvnd']
        }

        # Check environment variables for driver paths that need directory binding
        # (Binding the full directory handles scanning/plugins better than individual files)
        driver_env_vars = {
            'libibverbs.so': ['IBVERBS_DRIVER_PATH']
        }

        for lib in getattr(parameters, 'libraries', []):
            # Check for driver directories relative to lib (Standard convention)
            if lib.name.startswith('libibverbs.so'):
                verbs_dir = lib.parent / 'libibverbs'
                if verbs_dir.exists():
                    parameters.files.add(verbs_dir)
            
            # Check for environment-defined driver paths
            for prefix, var_names in driver_env_vars.items():
                if lib.name.startswith(prefix):
                    for var in var_names:
                        if var in os.environ:
                            path = Path(os.environ[var])
                            if path.exists():
                                parameters.files.add(path)

            # Check for configuration directories to mask
            for prefix, subdirs in config_masks.items():
                if lib.name.startswith(prefix):
                    for conf_root in system_config_dirs:
                        for subdir in subdirs:
                            conf_path = Path(conf_root) / subdir
                            if conf_path.exists():
                                parameters.files.add(conf_path)


        execute_command = _format_execute(parameters)

        def _supports_end_of_options(tokens: List[str]) -> bool:
            if not tokens:
                return False
            name = Path(tokens[0]).name
            return name in {"mpirun", "mpiexec", "srun", "jsrun", "aprun", "ibrun"}

        launcher_tokens = list(launcher)
        # Insert end-of-options delimiter if not already present
        if _supports_end_of_options(launcher_tokens) and (not launcher_tokens or launcher_tokens[-1] != "--"):
            launcher_tokens.append("--")

        full_command = [*launcher_tokens, *wi4mpi_call, *execute_command, *program]

        if variables.is_dry_run():
            print(' '.join(map(str, full_command)))
            return EXIT_SUCCESS

        retval, _ = run_e4scl_subprocess(full_command)
        return retval


SUMMARY = "Launch a process in a container with a tailored environment."
COMMAND = LaunchCommand(__name__, summary_fmt=SUMMARY)



"""
Library analysis and manipulation helpers
"""

from functools import lru_cache

# Symbols imported for ease of use
# pylint: disable=W0611
from sotools import is_elf, library_links
# pylint: disable=W0611
from sotools.ldd import ldd
# pylint: disable=W0611
from sotools.linker import resolve
# pylint: disable=W0611
from sotools.libraryset import LibrarySet, Library
# pylint: disable=W0611
from sotools.dl_cache import cache_libraries

from e4s_cl.error import InternalError
from e4s_cl.logger import get_logger
from e4s_cl.cf.version import Version
from e4s_cl.util import JSON_HOOKS

LOGGER = get_logger(__name__)


@lru_cache()
def libc_version():
    """
    -> e4s_cl.cf.version.Version
    Get the version number of the libc available on the host
    Caches the result
    """
    path = resolve('libc.so.6')
    if not path:
        raise InternalError("libc not found on host")

    data = Library.from_path(path)

    # Get the version with major 2 from the defined versions,
    # as almost all libc implementations have the GLIBC_3.4 symbol
    libc_ver = max(
        filter(lambda x: x and x.major == 2,
               [Version(s) for s in data.defined_versions]))

    return libc_ver



"""Management of the e4s-cl configuration file."""
import yaml
from pathlib import Path
from e4s_cl import EXIT_SUCCESS, EXIT_FAILURE, logger
from e4s_cl.cli import arguments
from e4s_cl.cli.command import AbstractCommand
from e4s_cl.config import ALLOWED_CONFIG, USER_CONFIG_PATH
from e4s_cl.util import mkdirp

LOGGER = logger.get_logger(__name__)


def _update_config(config_file, key, value):
    config = {}
    if config_file and Path(config_file).exists():
        with open(config_file, 'r', encoding='utf-8') as config_fh:
            config = yaml.safe_load(config_fh) or {}

    flattened_key = "_".join(key.split('.'))
    allowed_keys = {f.key: f for f in ALLOWED_CONFIG.flatten()}

    if flattened_key not in allowed_keys:
        LOGGER.error(
            "Invalid configuration key '%s'. Available keys: %s",
            key, ', '.join(allowed_keys.keys()))
        return False

    field_info = allowed_keys[flattened_key]

    try:
        if field_info.expected_type is bool:
            if value.lower() in ['true', 'yes', '1']:
                cast_value = True
            elif value.lower() in ['false', 'no', '0']:
                cast_value = False
            else:
                raise ValueError("Not a boolean")
        else:
            cast_value = field_info.expected_type(value)
    except ValueError:
        LOGGER.error(
            "Invalid value '%s' for key '%s'. Expected type: %s",
            value, key, field_info.expected_type.__name__)
        return False

    def set_recursive(current_dict, key_parts, val):
        if not key_parts:
            return
        head = key_parts[0]
        if len(key_parts) == 1:
            current_dict[head] = val
        else:
            if head not in current_dict:
                current_dict[head] = {}
            if not isinstance(current_dict[head], dict):
                current_dict[head] = {}
            set_recursive(current_dict[head], key_parts[1:], val)

    def get_path_segments(group, flat_key):
        for field in group.fields:
            if hasattr(field, 'fields'):
                prefix = field.key + "_" if field.key else ""
                if flat_key.startswith(prefix):
                    suffix = flat_key[len(prefix):]
                    res = get_path_segments(field, suffix)
                    if res is not None:
                        return [field.key] + res if field.key else res
            else:
                if field.key == flat_key:
                    return [field.key]
        return None

    segments = get_path_segments(ALLOWED_CONFIG, flattened_key)

    if not segments:
        LOGGER.error("Could not resolve key '%s' to configuration structure.", key)
        return False

    set_recursive(config, segments, cast_value)

    if not mkdirp(Path(config_file).parent):
        LOGGER.error("Could not create directory for %s", config_file)
        return False

    with open(config_file, 'w', encoding='utf-8') as config_fh:
        yaml.safe_dump(config, config_fh, default_flow_style=False)

    return True


def unflatten(flat_dict):
    """Reconstruct a nested config dict from a flat keyed dict.

    For example ``'wi4mpi_install_directory'`` becomes
    ``{'wi4mpi': {'install_directory': ...}}``.  The ALLOWED_CONFIG tree is
    used to determine grouping.
    """
    nested = {}

    def set_recursive(current_dict, key_parts, val):
        if not key_parts:
            return
        head = key_parts[0]
        if len(key_parts) == 1:
            current_dict[head] = val
        else:
            if head not in current_dict:
                current_dict[head] = {}
            if not isinstance(current_dict[head], dict):  # pragma: no cover
                current_dict[head] = {}
            set_recursive(current_dict[head], key_parts[1:], val)

    key_path_map = {}

    def traverse(group, current_path):
        for field in group.fields:
            if hasattr(field, 'fields'):
                traverse(field, current_path + [field.key])
            else:
                full_path = current_path + [field.key]
                flat_key = "_".join(filter(None, full_path))
                key_path_map[flat_key] = full_path

    traverse(ALLOWED_CONFIG, [])

    for flat_key, value in flat_dict.items():
        if flat_key in key_path_map:
            path = key_path_map[flat_key]
            set_recursive(nested, path, value)
        else:
            nested[flat_key] = value

    return nested


class ConfigCommand(AbstractCommand):
    """Manage the configuration."""

    def _construct_parser(self):
        parser = arguments.get_parser(prog=self.command, description=self.summary)
        subparsers = parser.add_subparsers(dest='action', help='Action to perform')
        subparsers.add_parser('list', help='List configuration options')
        set_parser = subparsers.add_parser('set', help='Set a configuration option')
        set_parser.add_argument(
            'key',
            help='Configuration key (can use dot notation, e.g. wi4mpi.install_directory)')
        set_parser.add_argument('value', help='Value to set')
        get_parser = subparsers.add_parser('get', help='Get a configuration option value')
        get_parser.add_argument(
            'key',
            help='Configuration key (can use dot or underscore notation)')
        return parser

    def main(self, argv):
        args = self._parse_args(argv)
        if not args.action:
            print(self.parser.format_help())
            return EXIT_FAILURE

        if args.action == 'list':
            from e4s_cl.config import CONFIGURATION
            nested = unflatten(CONFIGURATION._fields)
            print(yaml.safe_dump(nested, default_flow_style=False))
            return EXIT_SUCCESS

        if args.action == 'set':
            key = args.key.replace('.', '_')
            if _update_config(USER_CONFIG_PATH, key, args.value):
                print(f"Updated {USER_CONFIG_PATH}")
                return EXIT_SUCCESS
            return EXIT_FAILURE

        if args.action == 'get':
            from e4s_cl.config import CONFIGURATION
            key = args.key.replace('.', '_')
            try:
                val = getattr(CONFIGURATION, key)
                print(val)
                return EXIT_SUCCESS
            except AttributeError:
                LOGGER.error("Key '%s' not found in configuration.", key)
                return EXIT_FAILURE

        return EXIT_SUCCESS


COMMAND = ConfigCommand(__name__, summary_fmt="Configuration management")
